# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import buildsets
from buildbot.data import resultspec
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import endpoint
from buildbot.test.util import interfaces as util_interfaces
from buildbot.util import epoch2datetime

A_TIMESTAMP = 1341700729
A_TIMESTAMP_EPOCH = epoch2datetime(A_TIMESTAMP)
EARLIER = 1248529376
EARLIER_EPOCH = epoch2datetime(EARLIER)


class BuildsetEndpoint(endpoint.EndpointMixin, unittest.TestCase):

    endpointClass = buildsets.BuildsetEndpoint
    resourceTypeClass = buildsets.Buildset

    def setUp(self):
        self.setUpEndpoint()
        self.db.insertTestData([
            fakedb.Buildset(id=13, reason='because I said so'),
            fakedb.SourceStamp(id=92),
            fakedb.SourceStamp(id=93),
            fakedb.BuildsetSourceStamp(buildsetid=13, sourcestampid=92),
            fakedb.BuildsetSourceStamp(buildsetid=13, sourcestampid=93),

            fakedb.Buildset(id=14, reason='no sourcestamps'),
        ])

    def tearDown(self):
        self.tearDownEndpoint()

    @defer.inlineCallbacks
    def test_get_existing(self):
        buildset = yield self.callGet(('buildsets', 13))

        self.validateData(buildset)
        self.assertEqual(buildset['reason'], 'because I said so')

    @defer.inlineCallbacks
    def test_get_existing_no_sourcestamps(self):
        buildset = yield self.callGet(('buildsets', 14))

        self.validateData(buildset)
        self.assertEqual(buildset['sourcestamps'], [])

    @defer.inlineCallbacks
    def test_get_missing(self):
        buildset = yield self.callGet(('buildsets', 99))

        self.assertEqual(buildset, None)


class BuildsetsEndpoint(endpoint.EndpointMixin, unittest.TestCase):

    endpointClass = buildsets.BuildsetsEndpoint
    resourceTypeClass = buildsets.Buildset

    def setUp(self):
        self.setUpEndpoint()
        self.db.insertTestData([
            fakedb.SourceStamp(id=92),
            fakedb.Buildset(id=13, complete=True),
            fakedb.Buildset(id=14, complete=False),
            fakedb.BuildsetSourceStamp(buildsetid=13, sourcestampid=92),
            fakedb.BuildsetSourceStamp(buildsetid=14, sourcestampid=92),
        ])

    def tearDown(self):
        self.tearDownEndpoint()

    @defer.inlineCallbacks
    def test_get(self):
        buildsets = yield self.callGet(('buildsets',))

        self.validateData(buildsets[0])
        self.assertEqual(buildsets[0]['bsid'], 13)
        self.validateData(buildsets[1])
        self.assertEqual(buildsets[1]['bsid'], 14)

    @defer.inlineCallbacks
    def test_get_complete(self):
        f = resultspec.Filter('complete', 'eq', [True])
        buildsets = yield self.callGet(('buildsets',),
                         resultSpec=resultspec.ResultSpec(filters=[f]))

        self.assertEqual(len(buildsets), 1)
        self.validateData(buildsets[0])
        self.assertEqual(buildsets[0]['bsid'], 13)

    @defer.inlineCallbacks
    def test_get_incomplete(self):
        f = resultspec.Filter('complete', 'eq', [False])
        buildsets = yield self.callGet(('buildsets',),
                         resultSpec=resultspec.ResultSpec(filters=[f]))

        self.assertEqual(len(buildsets), 1)
        self.validateData(buildsets[0])
        self.assertEqual(buildsets[0]['bsid'], 14)


class Buildset(TestReactorMixin, util_interfaces.InterfaceTests,
               unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        self.master = fakemaster.make_master(self, wantMq=True, wantDb=True,
                                             wantData=True)
        self.rtype = buildsets.Buildset(self.master)
        return self.master.db.insertTestData([
            fakedb.SourceStamp(id=234, branch='br', codebase='cb',
                               project='pr', repository='rep', revision='rev',
                               created_at=89834834),
            fakedb.Builder(id=42, name='bldr1'),
            fakedb.Builder(id=43, name='bldr2'),
        ])

    SS234_DATA = {'branch': 'br', 'codebase': 'cb', 'patch': None,
                  'project': 'pr', 'repository': 'rep', 'revision': 'rev',
                  'created_at': epoch2datetime(89834834), 'ssid': 234}

    def test_signature_addBuildset(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.addBuildset,  # fake
            self.rtype.addBuildset)  # real
        def addBuildset(self, waited_for, scheduler=None, sourcestamps=None, reason='',
                        properties=None, builderids=None, external_idstring=None,
                        parent_buildid=None, parent_relationship=None):
            pass

    @defer.inlineCallbacks
    def do_test_addBuildset(self, kwargs, expectedReturn,
                            expectedMessages, expectedBuildset):
        """Run a test of addBuildset.

        @param kwargs: kwargs to addBuildset
        @param expectedReturn: expected return value - tuple of (bsid, brids)
        @param expectedMessages: expected mq messages transmitted
        @param expectedBuildset: expected buildset inserted into the db

        The buildset is added at time A_TIMESTAMP.
        Note that addBuildset does not add sourcestamps, so this method assumes
        there are none in the db.
        """
        self.reactor.advance(A_TIMESTAMP)

        (bsid, brids) = yield self.rtype.addBuildset(**kwargs)
        self.assertEqual((bsid, brids), expectedReturn)
        # check the correct message was received
        self.master.mq.assertProductions(
            expectedMessages, orderMatters=False)
        # and that the correct data was inserted into the db
        self.master.db.buildsets.assertBuildset(bsid, expectedBuildset)

    def _buildRequestMessageDict(self, brid, bsid, builderid):
        return {'builderid': builderid,
                'buildrequestid': brid,
                'buildsetid': bsid,
                'claimed': False,
                'claimed_at': None,
                'claimed_by_masterid': None,
                'complete': False,
                'complete_at': None,
                'priority': 0,
                'results': -1,
                'submitted_at': epoch2datetime(A_TIMESTAMP),
                'waited_for': True,
                'properties': None}

    def _buildRequestMessage1(self, brid, bsid, builderid):
        return (
            ('buildsets', str(bsid),
             'builders', str(builderid),
             'buildrequests', str(brid), 'new'),
            self._buildRequestMessageDict(brid, bsid, builderid))

    def _buildRequestMessage2(self, brid, bsid, builderid):
        return (
            ('buildrequests', str(brid), 'new'),
            self._buildRequestMessageDict(brid, bsid, builderid))

    def _buildRequestMessage3(self, brid, bsid, builderid):
        return (
            ('builders', str(builderid),
             'buildrequests', str(brid), 'new'),
            self._buildRequestMessageDict(brid, bsid, builderid))

    def _buildsetMessage(self, bsid, external_idstring='extid',
                         reason='because', scheduler='fakesched', sourcestampids=None,
                         submitted_at=A_TIMESTAMP):
        if sourcestampids is None:
            sourcestampids = [234]
        ssmap = {234: self.SS234_DATA}
        return (
            ('buildsets', str(bsid), 'new'),
            dict(bsid=bsid, complete=False, complete_at=None,
                 external_idstring=external_idstring, reason=reason,
                 results=None, scheduler=scheduler,
                 sourcestamps=[ssmap[ssid] for ssid in sourcestampids],
                 submitted_at=submitted_at))

    def _buildsetCompleteMessage(self, bsid, complete_at=A_TIMESTAMP_EPOCH,
                                 submitted_at=A_TIMESTAMP_EPOCH, external_idstring='extid',
                                 reason='because', results=0, sourcestampids=None):
        if sourcestampids is None:
            sourcestampids = [234]
        ssmap = {234: self.SS234_DATA}
        return (
            ('buildsets', str(bsid), 'complete'),
            dict(bsid=bsid, complete=True, complete_at=complete_at,
                 external_idstring=external_idstring, reason=reason,
                 results=results, submitted_at=submitted_at,
                 sourcestamps=[ssmap[ssid] for ssid in sourcestampids]))

    def test_addBuildset_two_builderNames(self):
        kwargs = dict(scheduler='fakesched', reason='because',
                      sourcestamps=[234], external_idstring='extid',
                      builderids=[42, 43], waited_for=True)
        expectedReturn = (200, {42: 1000, 43: 1001})
        expectedMessages = [
            self._buildRequestMessage1(1000, 200, 42),
            self._buildRequestMessage2(1000, 200, 42),
            self._buildRequestMessage3(1000, 200, 42),
            self._buildRequestMessage1(1001, 200, 43),
            self._buildRequestMessage2(1001, 200, 43),
            self._buildRequestMessage3(1001, 200, 43),
            self._buildsetMessage(200),
        ]
        expectedBuildset = dict(reason='because',
                                properties={},
                                external_idstring='extid')
        return self.do_test_addBuildset(kwargs,
                                        expectedReturn, expectedMessages, expectedBuildset)

    def test_addBuildset_no_builderNames(self):
        kwargs = dict(scheduler='fakesched', reason='because',
                      sourcestamps=[234], external_idstring='extid', waited_for=False)
        expectedReturn = (200, {})
        expectedMessages = [
            self._buildsetMessage(200),
            # with no builderNames, this is done already
            self._buildsetCompleteMessage(200),
        ]
        expectedBuildset = dict(reason='because',
                                properties={},
                                external_idstring='extid')
        return self.do_test_addBuildset(kwargs,
                                        expectedReturn, expectedMessages, expectedBuildset)

    def test_signature_maybeBuildsetComplete(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.maybeBuildsetComplete,  # fake
            self.rtype.maybeBuildsetComplete)  # real
        def maybeBuildsetComplete(self, bsid):
            pass

    @defer.inlineCallbacks
    def do_test_maybeBuildsetComplete(self,
                                      buildRequestCompletions=None,
                                      buildRequestResults=None,
                                      buildsetComplete=False,
                                      expectComplete=False,
                                      expectMessage=False,
                                      expectSuccess=True):
        """Test maybeBuildsetComplete.

        @param buildRequestCompletions: dict mapping brid to True if complete,
            else False (and defaulting to False)
        @param buildRequestResults: dict mapping brid to result (defaulting
            to SUCCESS)
        @param buildsetComplete: true if the buildset is already complete
        @param expectComplete: true if the buildset should be complete at exit
        @param expectMessage: true if a buildset completion message is expected
        @param expectSuccess: if expectComplete, whether to expect the buildset
            to be complete

        This first adds two buildsets to the database - 72 and 73.  Buildset 72
        is already complete if buildsetComplete is true; 73 is not complete.
        It adds four buildrequests - 42, 43, and 44 for buildset 72, and 45 for
        buildset 73.  The completion and results are based on
        buidlRequestCompletions and buildRequestResults.

        Then, maybeBuildsetComplete is called for buildset 72, and the
        expectations are checked.
        """
        if buildRequestCompletions is None:
            buildRequestCompletions = {}
        if buildRequestResults is None:
            buildRequestResults = {}

        self.reactor.advance(A_TIMESTAMP)

        def mkbr(brid, bsid=72):
            return fakedb.BuildRequest(id=brid, buildsetid=bsid, builderid=42,
                                       complete=buildRequestCompletions.get(
                                           brid),
                                       results=buildRequestResults.get(brid, SUCCESS))
        yield self.master.db.insertTestData([
            fakedb.Builder(id=42, name='bldr1'),
            fakedb.Buildset(id=72,
                            submitted_at=EARLIER,
                            complete=buildsetComplete,
                            complete_at=A_TIMESTAMP if buildsetComplete else None),
            mkbr(42), mkbr(43), mkbr(44),
            fakedb.BuildsetSourceStamp(buildsetid=72, sourcestampid=234),
            fakedb.Buildset(id=73,
                            complete=False),
            mkbr(45, bsid=73),
            fakedb.BuildsetSourceStamp(buildsetid=73, sourcestampid=234),
        ])

        yield self.rtype.maybeBuildsetComplete(72)

        self.master.db.buildsets.assertBuildsetCompletion(72, expectComplete)
        if expectMessage:
            self.assertEqual(self.master.mq.productions, [
                self._buildsetCompleteMessage(72,
                                              results=SUCCESS if expectSuccess else FAILURE,
                                              submitted_at=EARLIER_EPOCH),
            ])
        else:
            self.assertEqual(self.master.mq.productions, [])

    def test_maybeBuildsetComplete_not_yet(self):
        # only brid 42 is complete, so the buildset is not complete
        return self.do_test_maybeBuildsetComplete(
            buildRequestCompletions={42: True})

    def test_maybeBuildsetComplete_complete(self):
        return self.do_test_maybeBuildsetComplete(
            buildRequestCompletions={42: True, 43: True, 44: True},
            expectComplete=True,
            expectMessage=True)

    def test_maybeBuildsetComplete_complete_failure(self):
        return self.do_test_maybeBuildsetComplete(
            buildRequestCompletions={42: True, 43: True, 44: True},
            buildRequestResults={43: FAILURE},
            expectComplete=True,
            expectMessage=True,
            expectSuccess=False)

    def test_maybeBuildsetComplete_already_complete(self):
        return self.do_test_maybeBuildsetComplete(
            buildRequestCompletions={42: True, 43: True, 44: True},
            buildsetComplete=True,
            expectComplete=True,
            expectMessage=False)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os

from twisted.python import util
from twisted.trial import unittest

from buildbot.config.master import FileLoader
from buildbot.scripts import runner
from buildbot.test.util import dirs
from buildbot.test.util.warnings import assertNotProducesWarnings
from buildbot.warnings import DeprecatedApiWarning


class RealConfigs(dirs.DirsMixin, unittest.TestCase):

    def setUp(self):
        self.setUpDirs('basedir')
        self.basedir = os.path.abspath('basedir')
        self.filename = os.path.abspath("test.cfg")

    def tearDown(self):
        self.tearDownDirs()

    def test_sample_config(self):
        filename = util.sibpath(runner.__file__, 'sample.cfg')
        with assertNotProducesWarnings(DeprecatedApiWarning):
            FileLoader(self.basedir, filename).loadConfig()

    def test_0_9_0b5_api_renamed_config(self):
        with open(self.filename, "w", encoding='utf-8') as f:
            f.write(sample_0_9_0b5_api_renamed)
        with assertNotProducesWarnings(DeprecatedApiWarning):
            FileLoader(self.basedir, self.filename).loadConfig()


# sample.cfg from various versions, with comments stripped.  Adjustments made
# for compatibility are marked with comments

# Template for master configuration just after worker renaming.
sample_0_9_0b5_api_renamed = """\
from buildbot.plugins import *

c = BuildmasterConfig = {}

c['workers'] = [worker.Worker("example-worker", "pass")]

c['protocols'] = {'pb': {'port': 9989}}

c['change_source'] = []
c['change_source'].append(changes.GitPoller(
        'https://github.com/buildbot/hello-world.git',
        workdir='gitpoller-workdir', branch='master',
        pollinterval=300))

c['schedulers'] = []
c['schedulers'].append(schedulers.SingleBranchScheduler(
                            name="all",
                            change_filter=util.ChangeFilter(branch='master'),
                            treeStableTimer=None,
                            builderNames=["runtests"]))
c['schedulers'].append(schedulers.ForceScheduler(
                            name="force",
                            builderNames=["runtests"]))

factory = util.BuildFactory()
factory.addStep(steps.Git(repourl='https://github.com/buildbot/hello-world.git', mode='incremental'))
factory.addStep(steps.ShellCommand(command=["trial", "hello"],
                                   env={"PYTHONPATH": "."}))

c['builders'] = []
c['builders'].append(
    util.BuilderConfig(name="runtests",
      workernames=["example-worker"],
      factory=factory))

c['title'] = "Pyflakes"
c['titleURL'] = "https://launchpad.net/pyflakes"

c['buildbotURL'] = "http://localhost:8010/"

c['www'] = dict(port=8010,
                plugins=dict(waterfall_view={}, console_view={}))

c['db'] = {
    'db_url' : "sqlite:///state.sqlite",
}
"""  # noqa pylint: disable=line-too-long



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import re

from buildbot.util import ComparableMixin
from buildbot.util import NotABranch


def extract_filter_values(values, filter_name):
    if not isinstance(values, (list, str)):
        raise ValueError(f"Values of filter {filter_name} must be list of strings or a string")
    if isinstance(values, str):
        values = [values]
    else:
        for value in values:
            if not isinstance(value, str):
                raise ValueError(f"Value of filter {filter_name} must be string")
    return values


def extract_filter_values_branch(values, filter_name):
    if not isinstance(values, (list, str, type(None))):
        raise ValueError(f"Values of filter {filter_name} must be list of strings, "
                         "a string or None")
    if isinstance(values, (str, type(None))):
        values = [values]
    else:
        for value in values:
            if not isinstance(value, (str, type(None))):
                raise ValueError(f"Value of filter {filter_name} must be string or None")
    return values


def extract_filter_values_regex(values, filter_name):
    if not isinstance(values, (list, str, re.Pattern)):
        raise ValueError(f"Values of filter {filter_name} must be list of strings, "
                         "a string or regex")
    if isinstance(values, (str, re.Pattern)):
        values = [values]
    else:
        for value in values:
            if not isinstance(value, (str, re.Pattern)):
                raise ValueError(f"Value of filter {filter_name} must be string or regex")
    return values


def extract_filter_values_dict(values, filter_name):
    if not isinstance(values, dict):
        raise ValueError(f"Value of filter {filter_name} must be dict")
    return {k: extract_filter_values(v, filter_name) for k, v in values.items()}


def extract_filter_values_dict_regex(values, filter_name):
    if not isinstance(values, dict):
        raise ValueError(f"Value of filter {filter_name} must be dict")
    return {k: extract_filter_values_regex(v, filter_name) for k, v in values.items()}


class _FilterExactMatch(ComparableMixin):
    compare_attrs = ('prop', 'values')

    def __init__(self, prop, values):
        self.prop = prop
        self.values = values

    def is_matched(self, value):
        return value in self.values

    def describe(self):
        return f'{self.prop} in {self.values}'


class _FilterExactMatchInverse(ComparableMixin):
    compare_attrs = ('prop', 'values')

    def __init__(self, prop, values):
        self.prop = prop
        self.values = values

    def is_matched(self, value):
        return value not in self.values

    def describe(self):
        return f'{self.prop} not in {self.values}'


class _FilterRegex(ComparableMixin):
    compare_attrs = ('prop', 'regexes')

    def __init__(self, prop, regexes):
        self.prop = prop
        self.regexes = [self._compile(regex) for regex in regexes]

    def _compile(self, regex):
        if isinstance(regex, re.Pattern):
            return regex
        return re.compile(regex)

    def is_matched(self, value):
        if value is None:
            return False
        for regex in self.regexes:
            if regex.match(value) is not None:
                return True
        return False

    def describe(self):
        return f'{self.prop} matches {self.regexes}'


class _FilterRegexInverse(ComparableMixin):
    compare_attrs = ('prop', 'regexes')

    def __init__(self, prop, regexes):
        self.prop = prop
        self.regexes = [self._compile(regex) for regex in regexes]

    def _compile(self, regex):
        if isinstance(regex, re.Pattern):
            return regex
        return re.compile(regex)

    def is_matched(self, value):
        if value is None:
            return True
        for regex in self.regexes:
            if regex.match(value) is not None:
                return False
        return True

    def describe(self):
        return f'{self.prop} does not match {self.regexes}'


def _create_branch_filters(eq, not_eq, regex, not_regex, prop):
    filters = []
    if eq is not NotABranch:
        values = extract_filter_values_branch(eq, prop + '_eq')
        filters.append(_FilterExactMatch(prop, values))

    if not_eq is not NotABranch:
        values = extract_filter_values_branch(not_eq, prop + '_not_eq')
        filters.append(_FilterExactMatchInverse(prop, values))

    if regex is not None:
        values = extract_filter_values_regex(regex, prop + '_re')
        filters.append(_FilterRegex(prop, values))

    if not_regex is not None:
        values = extract_filter_values_regex(not_regex, prop + '_not_re')
        filters.append(_FilterRegexInverse(prop, values))

    return filters


def _create_filters(eq, not_eq, regex, not_regex, prop):
    filters = []
    if eq is not None:
        values = extract_filter_values(eq, prop + '_eq')
        filters.append(_FilterExactMatch(prop, values))

    if not_eq is not None:
        values = extract_filter_values(not_eq, prop + '_not_eq')
        filters.append(_FilterExactMatchInverse(prop, values))

    if regex is not None:
        values = extract_filter_values_regex(regex, prop + '_re')
        filters.append(_FilterRegex(prop, values))

    if not_regex is not None:
        values = extract_filter_values_regex(not_regex, prop + '_not_re')
        filters.append(_FilterRegexInverse(prop, values))

    return filters


def _create_property_filters(eq, not_eq, regex, not_regex, arg_prefix):
    filters = []
    if eq is not None:
        values_dict = extract_filter_values_dict(eq, arg_prefix + '_eq')
        filters += [_FilterExactMatch(prop, values) for prop, values in values_dict.items()]

    if not_eq is not None:
        values_dict = extract_filter_values_dict(not_eq, arg_prefix + '_not_eq')
        filters += [_FilterExactMatchInverse(prop, values) for prop, values in values_dict.items()]

    if regex is not None:
        values_dict = extract_filter_values_dict_regex(regex, arg_prefix + '_re')
        filters += [_FilterRegex(prop, values) for prop, values in values_dict.items()]

    if not_regex is not None:
        values_dict = extract_filter_values_dict_regex(not_regex, arg_prefix + '_not_re')
        filters += [_FilterRegexInverse(prop, values) for prop, values in values_dict.items()]

    return filters


class SourceStampFilter(ComparableMixin):

    compare_attrs = (
        'filter_fn',
        'filters',
    )

    def __init__(self,
                 # gets a SourceStamp dictionary, returns boolean
                 filter_fn=None,

                 project_eq=None, project_not_eq=None, project_re=None, project_not_re=None,
                 repository_eq=None, repository_not_eq=None,
                 repository_re=None, repository_not_re=None,
                 branch_eq=NotABranch, branch_not_eq=NotABranch, branch_re=None, branch_not_re=None,
                 codebase_eq=None, codebase_not_eq=None, codebase_re=None, codebase_not_re=None):

        self.filter_fn = filter_fn
        self.filters = _create_filters(
            project_eq,
            project_not_eq,
            project_re,
            project_not_re,
            'project',
        )
        self.filters += _create_filters(
            codebase_eq,
            codebase_not_eq,
            codebase_re,
            codebase_not_re,
            'codebase',
        )
        self.filters += _create_filters(
            repository_eq,
            repository_not_eq,
            repository_re,
            repository_not_re,
            'repository',
        )
        self.filters += _create_branch_filters(
            branch_eq,
            branch_not_eq,
            branch_re,
            branch_not_re,
            'branch',
        )

    def is_matched(self, ss):
        if self.filter_fn is not None and not self.filter_fn(ss):
            return False
        for filter in self.filters:
            value = ss.get(filter.prop, '')
            if not filter.is_matched(value):
                return False
        return True

    def __repr__(self):
        filters = []
        if self.filter_fn is not None:
            filters.append(f'{self.filter_fn.__name__}()')
        filters += [filter.describe() for filter in self.filters]
        return f"<{self.__class__.__name__} on {' and '.join(filters)}>"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from io import BytesIO

from mock import Mock

from twisted.internet import defer
from twisted.web import server

from buildbot.test.fake import fakemaster


def fakeMasterForHooks(testcase):
    # testcase must derive from TestReactorMixin and setup_test_reactor()
    # must be called before calling this function.

    master = fakemaster.make_master(testcase, wantData=True)
    master.www = Mock()
    return master


class FakeRequest(Mock):

    """
    A fake Twisted Web Request object, including some pointers to the
    buildmaster and an addChange method on that master which will append its
    arguments to self.addedChanges.
    """

    written = b''
    finished = False
    redirected_to = None
    failure = None

    def __init__(self, args=None, content=b''):
        super().__init__()

        if args is None:
            args = {}

        self.args = args
        self.content = BytesIO(content)
        self.site = Mock()
        self.site.buildbot_service = Mock()
        self.uri = b'/'
        self.prepath = []
        self.method = b'GET'
        self.received_headers = {}

        self.deferred = defer.Deferred()

    def getHeader(self, key):
        return self.received_headers.get(key)

    def write(self, data):
        self.written = self.written + data

    def redirect(self, url):
        self.redirected_to = url

    def finish(self):
        self.finished = True
        self.deferred.callback(None)

    def processingFailed(self, f):
        self.deferred.errback(f)

    # work around http://code.google.com/p/mock/issues/detail?id=105
    def _get_child_mock(self, **kw):
        return Mock(**kw)

    # cribed from twisted.web.test._util._render
    def test_render(self, resource):
        for arg in self.args:
            if not isinstance(arg, bytes):
                raise ValueError(f"self.args: {self.args!r},  contains values which are not bytes")

        if self.uri and not isinstance(self.uri, bytes):
            raise ValueError(f"self.uri: {self.uri!r} is {type(self.uri)}, not bytes")

        if self.method and not isinstance(self.method, bytes):
            raise ValueError(f"self.method: {self.method!r} is {type(self.method)}, not bytes")

        result = resource.render(self)
        if isinstance(result, bytes):
            self.write(result)
            self.finish()
            return self.deferred
        elif isinstance(result, str):
            raise ValueError(f"{resource.render!r} should return bytes, not {type(result)}: "
                             f"{result!r}")
        elif result is server.NOT_DONE_YET:
            return self.deferred
        else:
            raise ValueError(f"Unexpected return value: {result!r}")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.internet import reactor
from twisted.trial import unittest

from buildbot.clients import sendchange as sendchange_client
from buildbot.scripts import sendchange
from buildbot.test.util import misc


class TestSendChange(misc.StdoutAssertionsMixin, unittest.TestCase):

    class FakeSender:

        def __init__(self, testcase, master, auth, encoding=None):
            self.master = master
            self.auth = auth
            self.encoding = encoding
            self.testcase = testcase

        def send(self, branch, revision, comments, files, **kwargs):
            kwargs['branch'] = branch
            kwargs['revision'] = revision
            kwargs['comments'] = comments
            kwargs['files'] = files
            self.send_kwargs = kwargs
            d = defer.Deferred()
            if self.testcase.fail:
                reactor.callLater(0, d.errback, RuntimeError("oh noes"))
            else:
                reactor.callLater(0, d.callback, None)
            return d

    def setUp(self):
        self.fail = False  # set to true to get Sender.send to fail

        def Sender_constr(*args, **kwargs):
            self.sender = self.FakeSender(self, *args, **kwargs)
            return self.sender
        self.patch(sendchange_client, 'Sender', Sender_constr)

        # undo the effects of @in_reactor
        self.patch(sendchange, 'sendchange', sendchange.sendchange._orig)

        self.setUpStdoutAssertions()

    @defer.inlineCallbacks
    def test_sendchange_config(self):
        rc = yield sendchange.sendchange(dict(encoding='utf16', who='me',
                                       auth=['a', 'b'], master='m', branch='br', category='cat',
                                       revision='rr', properties={'a': 'b'}, repository='rep',
                                       project='prj', vc='git', revlink='rl', when=1234.0,
                                       comments='comm', files=('a', 'b'), codebase='cb'))

        self.assertEqual((self.sender.master, self.sender.auth,
                          self.sender.encoding, self.sender.send_kwargs,
                          self.getStdout(), rc),
                         ('m', ['a', 'b'], 'utf16', {
                             'branch': 'br',
                             'category': 'cat',
                             'codebase': 'cb',
                             'comments': 'comm',
                             'files': ('a', 'b'),
                             'project': 'prj',
                             'properties': {'a': 'b'},
                             'repository': 'rep',
                             'revision': 'rr',
                             'revlink': 'rl',
                             'when': 1234.0,
                             'who': 'me',
                             'vc': 'git'},
                             'change sent successfully', 0))

    @defer.inlineCallbacks
    def test_sendchange_config_no_codebase(self):
        rc = yield sendchange.sendchange(dict(encoding='utf16', who='me',
                                       auth=['a', 'b'], master='m', branch='br', category='cat',
                                       revision='rr', properties={'a': 'b'}, repository='rep',
                                       project='prj', vc='git', revlink='rl', when=1234.0,
                                       comments='comm', files=('a', 'b')))

        self.assertEqual((self.sender.master, self.sender.auth,
                          self.sender.encoding, self.sender.send_kwargs,
                          self.getStdout(), rc),
                         ('m', ['a', 'b'], 'utf16', {
                             'branch': 'br',
                             'category': 'cat',
                             'codebase': None,
                             'comments': 'comm',
                             'files': ('a', 'b'),
                             'project': 'prj',
                             'properties': {'a': 'b'},
                             'repository': 'rep',
                             'revision': 'rr',
                             'revlink': 'rl',
                             'when': 1234.0,
                             'who': 'me',
                             'vc': 'git'},
                             'change sent successfully', 0))

    @defer.inlineCallbacks
    def test_sendchange_fail(self):
        self.fail = True
        rc = yield sendchange.sendchange({})

        self.assertEqual((self.getStdout().split('\n')[0], rc),
                         ('change not sent:', 1))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.python import failure
from twisted.trial import unittest

from buildbot.util import subscription


class TestException(Exception):
    pass


class subscriptions(unittest.TestCase):

    def setUp(self):
        self.subpt = subscription.SubscriptionPoint('test_sub')

    def test_str(self):
        self.assertIn('test_sub', str(self.subpt))

    def test_subscribe_unsubscribe(self):
        state = []

        def cb(*args, **kwargs):
            state.append((args, kwargs))

        # subscribe
        sub = self.subpt.subscribe(cb)
        self.assertTrue(isinstance(sub, subscription.Subscription))
        self.assertEqual(state, [])

        # deliver
        self.subpt.deliver(1, 2, a=3, b=4)
        self.assertEqual(state, [((1, 2), dict(a=3, b=4))])
        state.pop()

        # unsubscribe
        sub.unsubscribe()

        # don't receive events anymore
        self.subpt.deliver(3, 4)
        self.assertEqual(state, [])

    def test_exception(self):
        def cb(*args, **kwargs):
            raise RuntimeError('mah bucket!')

        self.subpt.subscribe(cb)
        self.subpt.deliver()  # should not raise

        exceptions = self.subpt.pop_exceptions()
        self.assertEqual(len(exceptions), 1)
        self.assertIsInstance(exceptions[0], RuntimeError)

        # log.err will cause Trial to complain about this error anyway, unless
        # we clean it up
        self.assertEqual(1, len(self.flushLoggedErrors(RuntimeError)))

    def test_deferred_exception(self):

        d = defer.Deferred()

        @defer.inlineCallbacks
        def cb_deferred(*args, **kwargs):
            yield d
            raise RuntimeError('msg')

        self.subpt.subscribe(cb_deferred)
        self.subpt.deliver()

        d.callback(None)

        exceptions = self.subpt.pop_exceptions()
        self.assertEqual(len(exceptions), 1)
        self.assertIsInstance(exceptions[0], failure.Failure)

        self.assertEqual(1, len(self.flushLoggedErrors(RuntimeError)))

    def test_deferred_exception_after_pop_exceptions(self):
        # waitForDeliveriesToFinish is forgotten to be called and exception happens after
        # pop_exceptions.
        d = defer.Deferred()

        @defer.inlineCallbacks
        def cb_deferred(*args, **kwargs):
            yield d
            raise TestException('msg')

        self.subpt.subscribe(cb_deferred)
        self.subpt.deliver()

        exceptions = self.subpt.pop_exceptions()

        d.callback(None)

        self.assertEqual(len(exceptions), 0)
        self.assertEqual(2, len(self.flushLoggedErrors(TestException)))

    def test_multiple_exceptions(self):

        d = defer.Deferred()

        @defer.inlineCallbacks
        def cb_deferred(*args, **kwargs):
            yield d
            raise RuntimeError('msg')

        def cb(*args, **kwargs):
            raise RuntimeError('msg')

        self.subpt.subscribe(cb_deferred)
        self.subpt.subscribe(cb)
        self.subpt.deliver()

        d.callback(None)

        exceptions = self.subpt.pop_exceptions()
        self.assertEqual(len(exceptions), 2)
        self.assertIsInstance(exceptions[0], RuntimeError)
        self.assertIsInstance(exceptions[1], failure.Failure)

        self.assertEqual(2, len(self.flushLoggedErrors(RuntimeError)))

    def test_deliveries_finished(self):
        state = []

        def create_cb(d):
            def cb(*args):
                state.append(args)
                return d
            return cb

        d1 = defer.Deferred()
        d2 = defer.Deferred()
        self.subpt.subscribe(create_cb(d1))
        self.subpt.subscribe(create_cb(d2))
        self.assertEqual(state, [])

        self.subpt.deliver(1, 2)
        self.assertEqual(state, [(1, 2), (1, 2)])

        d = self.subpt.waitForDeliveriesToFinish()
        self.assertFalse(d.called)

        d1.callback(None)
        self.assertFalse(d.called)

        d2.callback(None)
        self.assertTrue(d.called)

        # when there are no waiting deliveries, should call the callback immediately
        d = self.subpt.waitForDeliveriesToFinish()
        self.assertTrue(d.called)

    def test_deliveries_not_finished_within_callback(self):
        state = []

        def cb(*args):
            state.append(args)
            d = self.subpt.waitForDeliveriesToFinish()
            self.assertFalse(d.called)

        self.subpt.subscribe(cb)
        self.assertEqual(state, [])

        self.subpt.deliver(1, 2)
        self.assertEqual(state, [(1, 2)])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.python import log

from buildbot.reporters.base import ReporterBase
from buildbot.reporters.generators.build import BuildStatusGenerator
from buildbot.reporters.message import MessageFormatterFunction
from buildbot.util import httpclientservice


class HttpStatusPush(ReporterBase):
    name = "HttpStatusPush"
    secrets = ["auth"]

    def checkConfig(self, serverUrl, auth=None, headers=None,
                    debug=None, verify=None, generators=None, **kwargs):

        if generators is None:
            generators = self._create_default_generators()

        super().checkConfig(generators=generators, **kwargs)
        httpclientservice.HTTPClientService.checkAvailable(self.__class__.__name__)

    @defer.inlineCallbacks
    def reconfigService(self, serverUrl, auth=None, headers=None,
                        debug=None, verify=None, generators=None,
                        **kwargs):
        self.debug = debug
        self.verify = verify

        if generators is None:
            generators = self._create_default_generators()

        yield super().reconfigService(generators=generators, **kwargs)

        self._http = yield httpclientservice.HTTPClientService.getService(
            self.master, serverUrl, auth=auth, headers=headers,
            debug=self.debug, verify=self.verify)

    def _create_default_generators(self):
        formatter = MessageFormatterFunction(lambda context: context['build'], 'json')
        return [
            BuildStatusGenerator(message_formatter=formatter, report_new=True)
        ]

    def is_status_2xx(self, code):
        return code // 100 == 2

    @defer.inlineCallbacks
    def sendMessage(self, reports):
        response = yield self._http.post("", json=reports[0]['body'])
        if not self.is_status_2xx(response.code):
            log.msg(f"{response.code}: unable to upload status: {response.content}")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import contextlib


class ConfigErrors(Exception):

    def __init__(self, errors=None):
        if errors is None:
            errors = []
        self.errors = errors[:]

    def __str__(self):
        return "\n".join(self.errors)

    def addError(self, msg):
        self.errors.append(msg)

    def merge(self, errors):
        self.errors.extend(errors.errors)

    def __bool__(self):
        return bool(len(self.errors))


_errors = None


def error(error, always_raise=False):
    if _errors is not None and not always_raise:
        _errors.addError(error)
    else:
        raise ConfigErrors([error])


@contextlib.contextmanager
def capture_config_errors(raise_on_error=False):
    global _errors
    prev_errors = _errors
    _errors = errors = ConfigErrors()
    try:
        yield errors
    except ConfigErrors as e:
        errors.merge(e)
    finally:
        _errors = prev_errors

    if raise_on_error and errors:
        raise errors



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import base64

from twisted.internet import defer

from buildbot.reporters.generators.build import BuildStatusGenerator
from buildbot.reporters.generators.buildset import BuildSetStatusGenerator
from buildbot.reporters.generators.worker import WorkerMissingGenerator
from buildbot.reporters.mail import ESMTPSenderFactory
from buildbot.reporters.mail import MailNotifier
from buildbot.reporters.message import MessageFormatter
from buildbot.reporters.message import MessageFormatterMissingWorker
from buildbot.reporters.pushover import PushoverNotifier
from buildbot.test.util.integration import RunMasterBase
from buildbot.util import bytes2unicode
from buildbot.util import unicode2bytes


# This integration test creates a master and worker environment,
# with one builders and a shellcommand step, and a MailNotifier
class NotifierMaster(RunMasterBase):

    if not ESMTPSenderFactory:
        skip = ("twisted-mail unavailable, "
                "see: https://twistedmatrix.com/trac/ticket/8770")

    @defer.inlineCallbacks
    def create_master_config(self, build_set_summary=False):
        self.mailDeferred = defer.Deferred()

        # patch MailNotifier.sendmail to know when the mail has been sent
        def sendMail(_, mail, recipients):
            self.mailDeferred.callback((mail.as_string(), recipients))
        self.patch(MailNotifier, "sendMail", sendMail)

        self.notification = defer.Deferred()

        def sendNotification(_, params):
            self.notification.callback(params)
        self.patch(PushoverNotifier, "sendNotification", sendNotification)

        yield self.setupConfig(masterConfig(build_set_summary=build_set_summary))

    @defer.inlineCallbacks
    def doTest(self, what):
        change = dict(branch="master",
                      files=["foo.c"],
                      author="author@foo.com",
                      committer="me@foo.com",
                      comments="good stuff",
                      revision="HEAD",
                      project="projectname"
                      )
        build = yield self.doForceBuild(wantSteps=True, useChange=change, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        mail, recipients = yield self.mailDeferred
        self.assertEqual(recipients, ["author@foo.com"])
        self.assertIn("From: bot@foo.com", mail)
        self.assertIn(f"Subject: =?utf-8?q?=E2=98=BA_Buildbot_=28Buildbot=29=3A_{what}_-_build_successful_=28master=29?=\n",  # noqa pylint: disable=line-too-long
                      mail)
        self.assertEncodedIn("A passing build has been detected on builder testy while", mail)
        params = yield self.notification
        self.assertEqual(build['buildid'], 1)
        self.assertEqual(params, {
            'title': f"☺ Buildbot (Buildbot): {what} - build successful (master)",
            'message': "This is a message."
        })

    def assertEncodedIn(self, text, mail):
        # python 2.6 default transfer in base64 for utf-8
        if "base64" not in mail:
            self.assertIn(text, mail)
        else:  # b64encode and remove '=' padding (hence [:-1])
            encodedBytes = base64.b64encode(unicode2bytes(text)).rstrip(b"=")
            encodedText = bytes2unicode(encodedBytes)
            self.assertIn(encodedText, mail)

    @defer.inlineCallbacks
    def test_notifiy_for_build(self):
        yield self.create_master_config(build_set_summary=False)
        yield self.doTest('testy')

    @defer.inlineCallbacks
    def test_notifiy_for_buildset(self):
        yield self.create_master_config(build_set_summary=True)
        yield self.doTest('projectname')

    @defer.inlineCallbacks
    def test_missing_worker(self):
        yield self.create_master_config(build_set_summary=False)
        yield self.master.data.updates.workerMissing(
            workerid='local1',
            masterid=self.master.masterid,
            last_connection='long time ago',
            notify=['admin@worker.org'],
        )
        mail, recipients = yield self.mailDeferred
        self.assertIn("From: bot@foo.com", mail)
        self.assertEqual(recipients, ['admin@worker.org'])
        self.assertIn("Subject: Buildbot Buildbot worker local1 missing", mail)
        self.assertIn("disconnected at long time ago", mail)
        self.assertEncodedIn("worker named local1 went away", mail)
        params = yield self.notification
        self.assertEqual(params, {'title': "Buildbot Buildbot worker local1 missing",
                                  'message': b"No worker."})


# master configuration
def masterConfig(build_set_summary):
    c = {}
    from buildbot.config import BuilderConfig
    from buildbot.process.factory import BuildFactory
    from buildbot.plugins import steps, schedulers, reporters
    c['schedulers'] = [
        schedulers.AnyBranchScheduler(
            name="sched",
            builderNames=["testy"])
    ]
    f = BuildFactory()
    f.addStep(steps.ShellCommand(command='echo hello'))
    c['builders'] = [
        BuilderConfig(name="testy",
                      workernames=["local1"],
                      factory=f)
    ]

    formatter = MessageFormatter(template='This is a message.')
    formatter_worker = MessageFormatterMissingWorker(template='No worker.')

    if build_set_summary:
        generators_mail = [
            BuildSetStatusGenerator(mode='all'),
            WorkerMissingGenerator(workers='all'),
        ]
        generators_pushover = [
            BuildSetStatusGenerator(mode='all', message_formatter=formatter),
            WorkerMissingGenerator(workers=['local1'], message_formatter=formatter_worker),
        ]
    else:
        generators_mail = [
            BuildStatusGenerator(mode='all'),
            WorkerMissingGenerator(workers='all'),
        ]
        generators_pushover = [
            BuildStatusGenerator(mode='all', message_formatter=formatter),
            WorkerMissingGenerator(workers=['local1'], message_formatter=formatter_worker),
        ]

    c['services'] = [
        reporters.MailNotifier("bot@foo.com", generators=generators_mail),
        reporters.PushoverNotifier('1234', 'abcd', generators=generators_pushover)
    ]
    return c



# -*- coding: utf-8 -*-
#
# Buildbot documentation build configuration file, created by
# sphinx-quickstart on Tue Aug 10 15:13:31 2010.
#
# This file is exec()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.


import os
import pkg_resources
import sys
import textwrap

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(1, os.path.dirname(os.path.abspath(__file__)))

try:
    from buildbot.util.raml import RamlSpec
    from buildbot.reporters.telegram import TelegramContact
except ImportError:
    sys.path.insert(2, os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                    os.pardir))
    from buildbot.util.raml import RamlSpec
    from buildbot.reporters.telegram import TelegramContact

# -- General configuration -----------------------------------------------
try:
    pkg_resources.require('docutils>=0.8')
except pkg_resources.ResolutionError as e:
    raise RuntimeError("docutils is not installed or has incompatible version. "
                       "Please install documentation dependencies with `pip "
                       "install buildbot[docs]`") from e
# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '4.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinx.ext.extlinks',
    'bbdocs.ext',
    'bbdocs.api_index',
    'sphinx_jinja',
    'sphinx_rtd_theme',
]
todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Buildbot'
copyright = 'Buildbot Team Members'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.

if 'VERSION' in os.environ:
    version = os.environ['VERSION']
else:
    gl = {'__file__': '../buildbot/__init__.py'}
    with open('../buildbot/__init__.py') as f:
        exec(f.read(), gl)  # pylint: disable=exec-used
    version = gl['version']

# The full version, including alpha/beta/rc tags.
release = version

# add a loud note about python 2
rst_prolog = textwrap.dedent("""\
.. caution:: Buildbot no longer supports Python 2.7 on the Buildbot master.
""")

# add a loud note for anyone looking at the latest docs
if release == 'latest':
    rst_prolog += textwrap.dedent("""\
    .. caution:: This page documents the latest, unreleased version of
        Buildbot.  For documentation for released versions, see
        http://docs.buildbot.net/current/.

    """)

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', 'release-notes/*.rst']

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = False

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'trac'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

intersphinx_mapping = {
    'python': ('https://python.readthedocs.io/en/latest/', None),
    'sqlalchemy': ('https://sqlalchemy.readthedocs.io/en/latest/', None),
}

extlinks = {
    'pull': ('https://github.com/buildbot/buildbot/pull/%s', 'pull request %s'),
    'issue': ('https://github.com/buildbot/buildbot/issues/%s', 'issue #%s'),
    # deprecated. Use issue instead, and point to Github
    'bug': ('http://trac.buildbot.net/ticket/%s', 'bug #%s'),
    # Renders as link with whole url, e.g.
    #   :src-link:`master`
    # renders as
    #   "https://github.com/buildbot/buildbot/blob/master/master".
    # Explicit title can be used for customizing how link looks like:
    #   :src-link:`master's directory <master>`
    'src-link': ('https://github.com/buildbot/buildbot/tree/master/%s', None),
    # "pretty" reference that looks like relative path in Buildbot source tree
    # by default.
    'src': ('https://github.com/buildbot/buildbot/tree/master/%s', '%s'),
    'contrib-src': ('https://github.com/buildbot/buildbot-contrib/tree/master/%s', '%s'),
}

# Sphinx' link checker.
linkcheck_ignore = [
    # Local URLs:
    r'^http://localhost.*',
    # Available only to logged-in users:
    r'^https://github\.com/settings/applications$',
    # Sites which uses SSL that Python 2 can't handle:
    r'^https://opensource\.org/licenses/gpl-2.0\.php$',
    r'^https://docs\.docker\.com/engine/installation/$',
    # Looks like server doesn't like user agent:
    r'^https://www\.microsoft\.com/en-us/download/details\.aspx\?id=17657$',
    # Example domain.
    r'^https?://(.+\.)?example\.org',
    # Anchor check fails on rendered user files on GitHub, since GitHub uses
    # custom prefix for anchors in user generated content.
    r'https://github\.com/buildbot/guanlecoja-ui/tree/master#changelog',
    r'http://mesosphere.github.io/marathon/docs/rest-api.html#post-v2-apps',
]
linkcheck_timeout = 10
linkcheck_retries = 3
linkcheck_workers = 20

# -- Options for HTML output ---------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {'stickysidebar': 'true'}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [
    '_themes'
]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = os.path.join('_images', 'full_logo.png')

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large or a png.
html_favicon = os.path.join('_static', 'icon.png')

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# We customize the rtd theme slightly
html_css_files = ['buildbot_rtd.css']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    '**': ['searchbox.html', 'localtoc.html', 'relations.html', 'sourcelink.html']
}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

html_use_index = True
html_use_modindex = False

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = 'Buildbotdoc'


# -- Options for LaTeX output --------------------------------------------

latex_elements = {}
# The paper size ('letter' or 'a4').
latex_elements['papersize'] = 'a4'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '11pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ('index', 'Buildbot.tex', 'Buildbot Documentation', 'Brian Warner', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = os.path.join('_images', 'header-text-transparent.png')

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# Three possible values for this option (see sphinx config manual) are:
# 1. 'no' - do not display URLs (default)
# 2. 'footnote' - display URLs in footnotes
# 3. 'inline' - display URLs inline in parentheses
latex_show_urls = 'inline'

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'buildbot', 'Buildbot Documentation', ['Brian Warner'], 1)
]

jinja_contexts = {
    "data_api": {'raml': RamlSpec()},
    "telegram": {'commands': TelegramContact.describe_commands()},
}

raml_spec = RamlSpec()
for raml_typename, raml_type in sorted(raml_spec.types.items()):
    jinja_contexts['data_api_' + raml_typename] = {
        'raml': raml_spec,
        'name': raml_typename,
        'type': raml_type,
    }

    doc_path = f'developer/raml/{raml_typename}.rst'
    if not os.path.exists(doc_path):
        raise Exception(f'File {doc_path} for RAML type {raml_typename} does not exist')

# Spell checker.
try:
    import enchant  # noqa # pylint: disable=unused-import
except ImportError as ex:
    print("enchant module import failed:\n"
          f"{ex}\n"
          "Spell checking disabled.",
          file=sys.stderr)

else:
    extensions.append('sphinxcontrib.spelling')
    spelling_show_suggestions = True



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.process.properties import Interpolate
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.reporters.github import HOSTED_BASE_URL
from buildbot.reporters.github import GitHubCommentPush
from buildbot.reporters.github import GitHubStatusPush
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.fake import httpclientservice as fakehttpclientservice
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util.config import ConfigErrorsMixin
from buildbot.test.util.reporter import ReporterTestMixin


class TestGitHubStatusPush(TestReactorMixin, unittest.TestCase, ConfigErrorsMixin,
                           ReporterTestMixin):

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()

        self.setup_reporter_test()
        # project must be in the form <owner>/<project>
        self.reporter_test_project = 'buildbot/buildbot'

        self.master = fakemaster.make_master(self, wantData=True, wantDb=True,
                                             wantMq=True)

        yield self.master.startService()
        self._http = yield fakehttpclientservice.HTTPClientService.getService(
            self.master, self,
            HOSTED_BASE_URL, headers={
                'Authorization': 'token XXYYZZ',
                'User-Agent': 'Buildbot'
            },
            debug=None, verify=None)
        self.sp = self.createService()
        yield self.sp.setServiceParent(self.master)

    def createService(self):
        return GitHubStatusPush(Interpolate('XXYYZZ'))

    def tearDown(self):
        return self.master.stopService()

    @defer.inlineCallbacks
    def test_basic(self):
        build = yield self.insert_build_new()
        # we make sure proper calls to txrequests have been made
        self._http.expect(
            'post',
            '/repos/buildbot/buildbot/statuses/d34db33fd43db33f',
            json={'state': 'pending',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build started.', 'context': 'buildbot/Builder0'})
        self._http.expect(
            'post',
            '/repos/buildbot/buildbot/statuses/d34db33fd43db33f',
            json={'state': 'success',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build done.', 'context': 'buildbot/Builder0'})
        self._http.expect(
            'post',
            '/repos/buildbot/buildbot/statuses/d34db33fd43db33f',
            json={'state': 'failure',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build done.', 'context': 'buildbot/Builder0'})

        build['complete'] = False
        build['results'] = None
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = FAILURE
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_empty(self):
        build = yield self.insert_build_new(insert_ss=False)
        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = FAILURE
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_source_stamp_no_props_nightly_scheduler(self):
        # no status updates are expected

        self.master.db.insertTestData([
            fakedb.Master(id=92),
            fakedb.Worker(id=13, name='wrk'),
            fakedb.Builder(id=79, name='Builder0'),
            fakedb.Buildset(id=98, results=SUCCESS, reason="test_reason1"),
            fakedb.BuildsetSourceStamp(buildsetid=98, sourcestampid=234),
            fakedb.SourceStamp(id=234, project=None, branch=None, revision=None,
                               repository=None, codebase=None),
            fakedb.BuildRequest(id=11, buildsetid=98, builderid=79),
            fakedb.Build(id=20, number=0, builderid=79, buildrequestid=11,
                         workerid=13, masterid=92, results=SUCCESS, state_string="build_text"),
            fakedb.BuildProperty(buildid=20, name="workername", value="wrk"),
            fakedb.BuildProperty(buildid=20, name="reason", value="because"),
            fakedb.BuildProperty(buildid=20, name="buildername", value="Builder0"),
            fakedb.BuildProperty(buildid=20, name="branch", value=None),
            fakedb.BuildProperty(buildid=20, name="codebase", value=""),
            fakedb.BuildProperty(buildid=20, name="project", value=""),
            fakedb.BuildProperty(buildid=20, name="repository", value=""),
            fakedb.BuildProperty(buildid=20, name="revision", value=None),
        ])

        self.setup_fake_get_changes_for_build(has_change=False)

        build = yield self.master.data.get(("builds", 20))

        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_multiple_source_stamps_no_props(self):
        repository = 'http://test_repo'
        project = 'test_user/test_project'
        codebase1 = 'test_codebase1'
        codebase2 = 'test_codebase2'
        codebase3 = 'test_codebase3'

        self._http.expect(
            'post',
            '/repos/test_user/test_project/statuses/rev1',
            json={'state': 'pending',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build started.', 'context': 'buildbot/Builder0'})
        self._http.expect(
            'post',
            '/repos/test_user/test_project/statuses/rev3',
            json={'state': 'pending',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build started.', 'context': 'buildbot/Builder0'})
        self._http.expect(
            'post',
            '/repos/test_user/test_project/statuses/rev1',
            json={'state': 'success',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build done.', 'context': 'buildbot/Builder0'})
        self._http.expect(
            'post',
            '/repos/test_user/test_project/statuses/rev3',
            json={'state': 'success',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build done.', 'context': 'buildbot/Builder0'})
        self._http.expect(
            'post',
            '/repos/test_user/test_project/statuses/rev1',
            json={'state': 'success',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build done.', 'context': 'buildbot/Builder0'})
        self._http.expect(
            'post',
            '/repos/test_user/test_project/statuses/rev3',
            json={'state': 'success',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build done.', 'context': 'buildbot/Builder0'})

        # note that the first sourcestamp only has revision, second only branch and only the third
        # has both
        self.master.db.insertTestData([
            fakedb.Master(id=92),
            fakedb.Worker(id=13, name='wrk'),
            fakedb.Builder(id=79, name='Builder0'),
            fakedb.Buildset(id=98, results=SUCCESS, reason="test_reason1"),
            fakedb.BuildsetSourceStamp(buildsetid=98, sourcestampid=234),
            fakedb.BuildsetSourceStamp(buildsetid=98, sourcestampid=235),
            fakedb.BuildsetSourceStamp(buildsetid=98, sourcestampid=236),
            fakedb.SourceStamp(id=234, project=project, branch=None, revision='rev1',
                               repository=repository, codebase=codebase1),
            fakedb.SourceStamp(id=235, project=project, branch='branch2', revision=None,
                               repository=repository, codebase=codebase2),
            fakedb.SourceStamp(id=236, project=project, branch='branch3', revision='rev3',
                               repository=repository, codebase=codebase3),
            fakedb.BuildRequest(id=11, buildsetid=98, builderid=79),
            fakedb.Build(id=20, number=0, builderid=79, buildrequestid=11,
                         workerid=13, masterid=92, results=SUCCESS, state_string="build_text"),
            fakedb.BuildProperty(buildid=20, name="buildername", value="Builder0"),
        ])

        self.setup_fake_get_changes_for_build(has_change=False)

        build = yield self.master.data.get(("builds", 20))

        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)


class TestGitHubStatusPushURL(TestReactorMixin, unittest.TestCase,
                              ReporterTestMixin):

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()

        self.setup_reporter_test()
        # project must be in the form <owner>/<project>
        self.reporter_test_project = 'buildbot'
        self.reporter_test_repo = 'https://github.com/buildbot1/buildbot1.git'

        self.master = fakemaster.make_master(self, wantData=True, wantDb=True,
                                             wantMq=True)

        yield self.master.startService()
        self._http = yield fakehttpclientservice.HTTPClientService.getService(
            self.master, self,
            HOSTED_BASE_URL, headers={
                'Authorization': 'token XXYYZZ',
                'User-Agent': 'Buildbot'
            },
            debug=None, verify=None)
        self.sp = self.createService()
        yield self.sp.setServiceParent(self.master)

    def createService(self):
        return GitHubStatusPush('XXYYZZ')

    def tearDown(self):
        return self.master.stopService()

    @defer.inlineCallbacks
    def test_ssh(self):
        self.reporter_test_repo = 'git@github.com:buildbot2/buildbot2.git'

        build = yield self.insert_build_new()
        # we make sure proper calls to txrequests have been made
        self._http.expect(
            'post',
            '/repos/buildbot2/buildbot2/statuses/d34db33fd43db33f',
            json={'state': 'pending',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build started.', 'context': 'buildbot/Builder0'})
        self._http.expect(
            'post',
            '/repos/buildbot2/buildbot2/statuses/d34db33fd43db33f',
            json={'state': 'success',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build done.', 'context': 'buildbot/Builder0'})
        self._http.expect(
            'post',
            '/repos/buildbot2/buildbot2/statuses/d34db33fd43db33f',
            json={'state': 'failure',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build done.', 'context': 'buildbot/Builder0'})

        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = FAILURE
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_https(self):
        build = yield self.insert_build_new()
        # we make sure proper calls to txrequests have been made
        self._http.expect(
            'post',
            '/repos/buildbot1/buildbot1/statuses/d34db33fd43db33f',
            json={'state': 'pending',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build started.', 'context': 'buildbot/Builder0'})
        self._http.expect(
            'post',
            '/repos/buildbot1/buildbot1/statuses/d34db33fd43db33f',
            json={'state': 'success',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build done.', 'context': 'buildbot/Builder0'})
        self._http.expect(
            'post',
            '/repos/buildbot1/buildbot1/statuses/d34db33fd43db33f',
            json={'state': 'failure',
                  'target_url': 'http://localhost:8080/#builders/79/builds/0',
                  'description': 'Build done.', 'context': 'buildbot/Builder0'})

        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = FAILURE
        yield self.sp._got_event(('builds', 20, 'finished'), build)


class TestGitHubCommentPush(TestGitHubStatusPush):

    def createService(self):
        return GitHubCommentPush('XXYYZZ')

    @defer.inlineCallbacks
    def test_basic(self):
        build = yield self.insert_build_new()
        # we make sure proper calls to txrequests have been made
        self._http.expect(
            'post',
            '/repos/buildbot/buildbot/issues/34/comments',
            json={'body': 'Build done.'})
        self._http.expect(
            'post',
            '/repos/buildbot/buildbot/issues/34/comments',
            json={'body': 'Build done.'})

        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = FAILURE
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_empty(self):
        build = yield self.insert_build_new(insert_ss=False)
        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = FAILURE
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_basic_branch_head(self):
        self.reporter_test_props['branch'] = 'refs/pull/13/head'
        build = yield self.insert_build_new()
        # we make sure proper calls to txrequests have been made
        self._http.expect(
            'post',
            '/repos/buildbot/buildbot/issues/13/comments',
            json={'body': 'Build done.'})
        self._http.expect(
            'post',
            '/repos/buildbot/buildbot/issues/13/comments',
            json={'body': 'Build done.'})

        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = FAILURE
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_multiple_source_stamps_no_props(self):
        repository = 'http://test_repo'
        project = 'test_user/test_project'
        codebase1 = 'test_codebase1'
        codebase2 = 'test_codebase2'
        codebase3 = 'test_codebase3'
        branch2 = 'refs/pull/4192/merge'
        branch3 = 'refs/pull/4193/merge'

        self._http.expect(
            'post',
            '/repos/test_user/test_project/issues/4192/comments',
            json={'body': 'Build done.'})
        self._http.expect(
            'post',
            '/repos/test_user/test_project/issues/4192/comments',
            json={'body': 'Build done.'})
        self._http.expect(
            'post',
            '/repos/test_user/test_project/issues/4192/comments',
            json={'body': 'Build done.'})
        self._http.expect(
            'post',
            '/repos/test_user/test_project/issues/4192/comments',
            json={'body': 'Build done.'})

        # note that the first sourcestamp only has revision, second only branch and only the third
        # has both
        self.master.db.insertTestData([
            fakedb.Master(id=92),
            fakedb.Worker(id=13, name='wrk'),
            fakedb.Builder(id=79, name='Builder0'),
            fakedb.Buildset(id=98, results=SUCCESS, reason="test_reason1"),
            fakedb.BuildsetSourceStamp(buildsetid=98, sourcestampid=234),
            fakedb.BuildsetSourceStamp(buildsetid=98, sourcestampid=235),
            fakedb.BuildsetSourceStamp(buildsetid=98, sourcestampid=236),
            fakedb.SourceStamp(id=234, project=project, branch=None, revision='rev1',
                               repository=repository, codebase=codebase1),
            fakedb.SourceStamp(id=235, project=project, branch=branch2, revision=None,
                               repository=repository, codebase=codebase2),
            fakedb.SourceStamp(id=236, project=project, branch=branch3, revision='rev3',
                               repository=repository, codebase=codebase3),
            fakedb.BuildRequest(id=11, buildsetid=98, builderid=79),
            fakedb.Build(id=20, number=0, builderid=79, buildrequestid=11,
                         workerid=13, masterid=92, results=SUCCESS, state_string="build_text"),
            fakedb.BuildProperty(buildid=20, name="buildername", value="Builder0"),
            fakedb.BuildProperty(buildid=20, name="branch", value=branch2),
        ])

        self.setup_fake_get_changes_for_build(has_change=False)

        build = yield self.master.data.get(("builds", 20))

        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = SUCCESS
        yield self.sp._got_event(('builds', 20, 'finished'), build)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test.fake import httpclientservice as fakehttpclientservice
from buildbot.util import httpclientservice
from buildbot.util import service


class myTestedService(service.BuildbotService):
    name = 'myTestedService'

    @defer.inlineCallbacks
    def reconfigService(self, baseurl):
        self._http = yield httpclientservice.HTTPClientService.getService(self.master, baseurl)

    @defer.inlineCallbacks
    def doGetRoot(self):
        res = yield self._http.get("/")
        # note that at this point, only the http response headers are received
        if res.code != 200:
            raise Exception(f"{res.code}: server did not succeed")
        res_json = yield res.json()
        # res.json() returns a deferred to represent the time needed to fetch the entire body
        return res_json


class Test(unittest.TestCase):

    @defer.inlineCallbacks
    def setUp(self):
        baseurl = 'http://127.0.0.1:8080'
        self.parent = service.MasterService()
        self._http = yield fakehttpclientservice.HTTPClientService.getService(
            self.parent, self, baseurl)
        self.tested = myTestedService(baseurl)

        yield self.tested.setServiceParent(self.parent)
        yield self.parent.startService()

    @defer.inlineCallbacks
    def test_root(self):
        self._http.expect("get", "/", content_json={'foo': 'bar'})

        response = yield self.tested.doGetRoot()
        self.assertEqual(response, {'foo': 'bar'})

    @defer.inlineCallbacks
    def test_root_error(self):
        self._http.expect("get", "/", content_json={'foo': 'bar'}, code=404)

        try:
            yield self.tested.doGetRoot()
        except Exception as e:
            self.assertEqual(str(e), '404: server did not succeed')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from xml.sax.saxutils import escape

import jinja2

from klein import Klein
from twisted.internet import defer

import cairocffi as cairo
import cairosvg
from buildbot.process.results import Results
from buildbot.util import bytes2unicode
from buildbot.www.plugin import Application


class Api:
    app = Klein()

    default = {  # note that these defaults are documented in configuration/www.rst
        "left_pad": 5,
        "left_text": "Build Status",
        "left_color": "#555",
        "right_pad": 5,
        "border_radius": 5,
        "style": "plastic",
        "template_name": "{style}.svg.j2",
        "font_face": "DejaVu Sans",
        "font_size": 11,
        "color_scheme": {
            "exception": "#007ec6",  # blue
            "failure": "#e05d44",    # red
            "retry": "#007ec6",      # blue
            "running": "#007ec6",    # blue
            "skipped": "a4a61d",     # yellowgreen
            "success": "#4c1",       # brightgreen
            "unknown": "#9f9f9f",    # lightgrey
            "warnings": "#dfb317"    # yellow
        }
    }

    def __init__(self, ep):
        self.ep = ep
        self.env = jinja2.Environment(loader=jinja2.ChoiceLoader([
            jinja2.PackageLoader('buildbot_badges'),
            jinja2.FileSystemLoader('templates')
        ]))

    def makeConfiguration(self, request):

        config = {}
        config.update(self.default)
        for k, v in self.ep.config.items():
            if k == 'color_scheme':
                config[k].update(v)
            else:
                config[k] = v

        for k, v in request.args.items():
            k = bytes2unicode(k)
            config[k] = escape(bytes2unicode(v[0]))
        return config

    @app.route("/<string:builder>.png", methods=['GET'])
    @defer.inlineCallbacks
    def getPng(self, request, builder):
        svg = yield self.getSvg(request, builder)
        request.setHeader('content-type', 'image/png')
        return cairosvg.svg2png(svg)

    @app.route("/<string:builder>.svg", methods=['GET'])
    @defer.inlineCallbacks
    def getSvg(self, request, builder):
        config = self.makeConfiguration(request)
        request.setHeader('content-type', 'image/svg+xml')
        request.setHeader('cache-control', 'no-cache')

        # get the last build for that builder using the data api
        last_build = yield self.ep.master.data.get(
            ("builders", builder, "builds"),
            limit=1, order=['-number'])

        # get the status text corresponding to results code
        results_txt = "unknown"
        if last_build:
            results = last_build[0]['results']
            complete = last_build[0]['complete']
            if not complete:
                results_txt = "running"
            elif results >= 0 and results < len(Results):
                results_txt = Results[results]

        svgdata = self.makesvg(results_txt, results_txt, left_text=config['left_text'],
                               config=config)
        return svgdata

    def textwidth(self, text, config):
        """Calculates the width of the specified text.
        """
        surface = cairo.SVGSurface(None, 1280, 200)
        ctx = cairo.Context(surface)
        ctx.select_font_face(config['font_face'],
                             cairo.FONT_SLANT_NORMAL,
                             cairo.FONT_WEIGHT_NORMAL)
        ctx.set_font_size(int(config['font_size']))
        return ctx.text_extents(text)[4]

    def makesvg(self, right_text, status=None, left_text=None,
                left_color=None, config=None):
        """Renders an SVG from the template, using the specified data
        """
        right_color = config['color_scheme'].get(status, "#9f9f9f")  # Grey

        left_text = left_text or config['left_text']
        left_color = left_color or config['left_color']

        left = {
            "color": left_color,
            "text": left_text,
            "width": self.textwidth(left_text, config)
        }
        right = {
            "color": right_color,
            "text": right_text,
            "width": self.textwidth(right_text, config)
        }

        template = self.env.get_template(config['template_name'].format(**config))
        return template.render(left=left, right=right, config=config)


# create the interface for the setuptools entry point
ep = Application(__name__, "Buildbot badges", ui=False)
ep.resource = Api(ep).app.resource()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer

from buildbot.test.fakedb.base import FakeDBComponent
from buildbot.test.fakedb.row import Row
from buildbot.test.util import validation


class Log(Row):
    table = "logs"

    id_column = 'id'
    required_columns = ('stepid', )

    def __init__(self, id=None, name='log29', slug='log29', stepid=None, complete=0, num_lines=0,
                 type='s'):
        super().__init__(id=id, name=name, slug=slug, stepid=stepid, complete=complete,
                         num_lines=num_lines, type=type)


class LogChunk(Row):
    table = "logchunks"

    required_columns = ('logid', )
    # 'content' column is sa.LargeBinary, it's bytestring.
    binary_columns = ('content',)

    def __init__(self, logid=None, first_line=0, last_line=0, content='', compressed=0):
        super().__init__(logid=logid, first_line=first_line, last_line=last_line, content=content,
                         compressed=compressed)


class FakeLogsComponent(FakeDBComponent):

    def setUp(self):
        self.logs = {}
        self.log_lines = {}  # { logid : [ lines ] }

    def insertTestData(self, rows):
        for row in rows:
            if isinstance(row, Log):
                self.logs[row.id] = row.values.copy()
        for row in rows:
            if isinstance(row, LogChunk):
                lines = self.log_lines.setdefault(row.logid, [])
                # make sure there are enough slots in the list
                if len(lines) < row.last_line + 1:
                    lines.append([None] * (row.last_line + 1 - len(lines)))
                row_lines = row.content.decode('utf-8').split('\n')
                lines[row.first_line:row.last_line + 1] = row_lines

    # component methods

    def _newId(self):
        id = 100
        while id in self.logs:
            id += 1
        return id

    def _row2dict(self, row):
        return dict(
            id=row['id'],
            stepid=row['stepid'],
            name=row['name'],
            slug=row['slug'],
            complete=bool(row['complete']),
            num_lines=row['num_lines'],
            type=row['type'])

    def getLog(self, logid):
        row = self.logs.get(logid)
        if not row:
            return defer.succeed(None)
        return defer.succeed(self._row2dict(row))

    def getLogBySlug(self, stepid, slug):
        row = None
        for row in self.logs.values():
            if row['slug'] == slug and row['stepid'] == stepid:
                break
        else:
            return defer.succeed(None)
        return defer.succeed(self._row2dict(row))

    def getLogs(self, stepid=None):
        return defer.succeed([
            self._row2dict(row)
            for row in self.logs.values()
            if row['stepid'] == stepid])

    def getLogLines(self, logid, first_line, last_line):
        if logid not in self.logs or first_line > last_line:
            return defer.succeed('')
        lines = self.log_lines.get(logid, [])
        rv = lines[first_line:last_line + 1]
        return defer.succeed('\n'.join(rv) + '\n' if rv else '')

    def addLog(self, stepid, name, slug, type):
        id = self._newId()
        self.logs[id] = dict(id=id, stepid=stepid,
                             name=name, slug=slug, type=type,
                             complete=0, num_lines=0)
        self.log_lines[id] = []
        return defer.succeed(id)

    def appendLog(self, logid, content):
        validation.verifyType(self.t, 'logid', logid,
                              validation.IntValidator())
        validation.verifyType(self.t, 'content', content,
                              validation.StringValidator())
        self.t.assertEqual(content[-1], '\n')
        content = content[:-1].split('\n')
        lines = self.log_lines[logid]
        lines.extend(content)
        num_lines = self.logs[logid]['num_lines'] = len(lines)
        return defer.succeed((num_lines - len(content), num_lines - 1))

    def finishLog(self, logid):
        if id in self.logs:
            self.logs['id'].complete = 1
        return defer.succeed(None)

    def compressLog(self, logid, force=False):
        return defer.succeed(None)

    def deleteOldLogChunks(self, older_than_timestamp):
        # not implemented
        self._deleted = older_than_timestamp
        return defer.succeed(1)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import masters
from buildbot.process.results import RETRY
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import endpoint
from buildbot.test.util import interfaces
from buildbot.util import epoch2datetime

SOMETIME = 1349016870
OTHERTIME = 1249016870


class MasterEndpoint(endpoint.EndpointMixin, unittest.TestCase):

    endpointClass = masters.MasterEndpoint
    resourceTypeClass = masters.Master

    def setUp(self):
        self.setUpEndpoint()
        self.master.name = "myname"
        self.db.insertTestData([
            fakedb.Master(id=13, name='some:master', active=False,
                          last_active=SOMETIME),
            fakedb.Master(id=14, name='other:master', active=False,
                          last_active=SOMETIME),
            fakedb.Builder(id=23, name='bldr1'),
            fakedb.BuilderMaster(builderid=23, masterid=13),
            fakedb.Builder(id=24, name='bldr2'),
        ])

    def tearDown(self):
        self.tearDownEndpoint()

    @defer.inlineCallbacks
    def test_get_existing(self):
        master = yield self.callGet(('masters', 14))

        self.validateData(master)
        self.assertEqual(master['name'], 'other:master')

    @defer.inlineCallbacks
    def test_get_builderid_existing(self):
        master = yield self.callGet(('builders', 23, 'masters', 13))

        self.validateData(master)
        self.assertEqual(master['name'], 'some:master')

    @defer.inlineCallbacks
    def test_get_builderid_no_match(self):
        master = yield self.callGet(('builders', 24, 'masters', 13))

        self.assertEqual(master, None)

    @defer.inlineCallbacks
    def test_get_builderid_missing(self):
        master = yield self.callGet(('builders', 25, 'masters', 13))

        self.assertEqual(master, None)

    @defer.inlineCallbacks
    def test_get_missing(self):
        master = yield self.callGet(('masters', 99))

        self.assertEqual(master, None)


class MastersEndpoint(endpoint.EndpointMixin, unittest.TestCase):

    endpointClass = masters.MastersEndpoint
    resourceTypeClass = masters.Master

    def setUp(self):
        self.setUpEndpoint()
        self.master.name = "myname"
        self.db.insertTestData([
            fakedb.Master(id=13, name='some:master', active=False,
                          last_active=SOMETIME),
            fakedb.Master(id=14, name='other:master', active=True,
                          last_active=OTHERTIME),
            fakedb.Builder(id=22),
            fakedb.BuilderMaster(masterid=13, builderid=22),
        ])

    def tearDown(self):
        self.tearDownEndpoint()

    @defer.inlineCallbacks
    def test_get(self):
        masters = yield self.callGet(('masters',))

        for m in masters:
            self.validateData(m)

        self.assertEqual(sorted([m['masterid'] for m in masters]),
                         [13, 14])

    @defer.inlineCallbacks
    def test_get_builderid(self):
        masters = yield self.callGet(('builders', 22, 'masters'))

        for m in masters:
            self.validateData(m)

        self.assertEqual(sorted([m['masterid'] for m in masters]),
                         [13])

    @defer.inlineCallbacks
    def test_get_builderid_missing(self):
        masters = yield self.callGet(('builders', 23, 'masters'))

        self.assertEqual(masters, [])


class Master(TestReactorMixin, interfaces.InterfaceTests, unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        self.master = fakemaster.make_master(self, wantMq=True, wantDb=True,
                                             wantData=True)
        self.rtype = masters.Master(self.master)

    def test_signature_masterActive(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.masterActive,  # fake
            self.rtype.masterActive)  # real
        def masterActive(self, name, masterid):
            pass

    @defer.inlineCallbacks
    def test_masterActive(self):
        self.reactor.advance(60)

        self.master.db.insertTestData([
            fakedb.Master(id=13, name='myname', active=0,
                          last_active=0),
            fakedb.Master(id=14, name='other', active=1,
                          last_active=0),
            fakedb.Master(id=15, name='other2', active=1,
                          last_active=0),
        ])

        # initial checkin
        yield self.rtype.masterActive(name='myname', masterid=13)
        master = yield self.master.db.masters.getMaster(13)
        self.assertEqual(master, dict(id=13, name='myname',
                                      active=True, last_active=epoch2datetime(60)))
        self.assertEqual(self.master.mq.productions, [
            (('masters', '13', 'started'),
             dict(masterid=13, name='myname', active=True)),
        ])
        self.master.mq.productions = []

        # updated checkin time, re-activation
        self.reactor.advance(60)
        yield self.master.db.masters.markMasterInactive(13)
        yield self.rtype.masterActive('myname', masterid=13)
        master = yield self.master.db.masters.getMaster(13)
        self.assertEqual(master, dict(id=13, name='myname',
                                      active=True, last_active=epoch2datetime(120)))
        self.assertEqual(self.master.mq.productions, [
            (('masters', '13', 'started'),
             dict(masterid=13, name='myname', active=True)),
        ])
        self.master.mq.productions = []

    def test_signature_masterStopped(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.masterStopped,  # fake
            self.rtype.masterStopped)  # real
        def masterStopped(self, name, masterid):
            pass

    @defer.inlineCallbacks
    def test_masterStopped(self):
        self.reactor.advance(60)

        self.master.db.insertTestData([
            fakedb.Master(id=13, name='aname', active=1,
                          last_active=self.reactor.seconds()),
        ])

        self.rtype._masterDeactivated = mock.Mock()
        yield self.rtype.masterStopped(name='aname', masterid=13)
        self.rtype._masterDeactivated. \
            assert_called_with(13, 'aname')

    @defer.inlineCallbacks
    def test_masterStopped_already(self):
        self.reactor.advance(60)

        self.master.db.insertTestData([
            fakedb.Master(id=13, name='aname', active=0,
                          last_active=0),
        ])

        self.rtype._masterDeactivated = mock.Mock()
        yield self.rtype.masterStopped(name='aname', masterid=13)
        self.rtype._masterDeactivated.assert_not_called()

    def test_signature_expireMasters(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.expireMasters,  # fake
            self.rtype.expireMasters)  # real
        def expireMasters(self, forceHouseKeeping=False):
            pass

    @defer.inlineCallbacks
    def test_expireMasters(self):
        self.reactor.advance(60)

        self.master.db.insertTestData([
            fakedb.Master(id=14, name='other', active=1,
                          last_active=0),
            fakedb.Master(id=15, name='other', active=1,
                          last_active=0),
        ])

        self.rtype._masterDeactivated = mock.Mock()

        # check after 10 minutes, and see #14 deactivated; #15 gets deactivated
        # by another master, so it's not included here
        self.reactor.advance(600)
        yield self.master.db.masters.markMasterInactive(15)
        yield self.rtype.expireMasters()
        master = yield self.master.db.masters.getMaster(14)
        self.assertEqual(master, dict(id=14, name='other',
                                      active=False,
                                      last_active=epoch2datetime(0)))
        self.rtype._masterDeactivated. \
            assert_called_with(14, 'other')

    @defer.inlineCallbacks
    def test_masterDeactivated(self):
        self.master.db.insertTestData([
            fakedb.Master(id=14, name='other', active=0,
                          last_active=0),

            # set up a running build with some steps
            fakedb.Builder(id=77, name='b1'),
            fakedb.Worker(id=13, name='wrk'),
            fakedb.Buildset(id=8822),
            fakedb.BuildRequest(id=82, builderid=77, buildsetid=8822),
            fakedb.BuildRequestClaim(brid=82, masterid=14,
                                     claimed_at=SOMETIME),
            fakedb.Build(id=13, builderid=77, masterid=14, workerid=13,
                         buildrequestid=82, number=3, results=None),
            fakedb.Step(id=200, buildid=13),
            fakedb.Log(id=2000, stepid=200, num_lines=2),
            fakedb.LogChunk(logid=2000, first_line=1, last_line=2,
                            content='ab\ncd')
        ])

        # mock out the _masterDeactivated methods this will call
        for rtype in 'builder', 'scheduler', 'changesource':
            rtype_obj = getattr(self.master.data.rtypes, rtype)
            m = mock.Mock(name=f'{rtype}._masterDeactivated',
                          spec=rtype_obj._masterDeactivated)
            m.side_effect = lambda masterid: defer.succeed(None)
            rtype_obj._masterDeactivated = m

        # and the update methods..
        for meth in 'finishBuild', 'finishStep', 'finishLog':
            m = mock.create_autospec(getattr(self.master.data.updates, meth))
            m.side_effect = lambda *args, **kwargs: defer.succeed(None)
            setattr(self.master.data.updates, meth, m)

        yield self.rtype._masterDeactivated(14, 'other')

        self.master.data.rtypes.builder._masterDeactivated. \
            assert_called_with(masterid=14)
        self.master.data.rtypes.scheduler._masterDeactivated. \
            assert_called_with(masterid=14)
        self.master.data.rtypes.changesource._masterDeactivated. \
            assert_called_with(masterid=14)

        # see that we finished off that build and its steps and logs
        updates = self.master.data.updates
        updates.finishLog.assert_called_with(logid=2000)
        updates.finishStep.assert_called_with(
            stepid=200, results=RETRY, hidden=False)
        updates.finishBuild.assert_called_with(buildid=13, results=RETRY)

        self.assertEqual(self.master.mq.productions, [
            (('masters', '14', 'stopped'),
             dict(masterid=14, name='other', active=False)),
        ])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
vault based providers
"""

from twisted.internet import defer

from buildbot import config
from buildbot.secrets.providers.base import SecretProviderBase
from buildbot.util import httpclientservice
from buildbot.warnings import warn_deprecated


class HashiCorpVaultSecretProvider(SecretProviderBase):
    """
    basic provider where each secret is stored in Vault KV secret engine
    """

    name = 'SecretInVault'

    def checkConfig(self, vaultServer=None, vaultToken=None, secretsmount=None,
                    apiVersion=1):
        warn_deprecated("3.4.0", "Use of HashiCorpVaultSecretProvider is deprecated and will be "
                        "removed in future releases. Use HashiCorpVaultKvSecretProvider instead")
        if not isinstance(vaultServer, str):
            config.error(f"vaultServer must be a string while it is {type(vaultServer)}")
        if not isinstance(vaultToken, str):
            config.error(f"vaultToken must be a string while it is {type(vaultToken)}")
        if apiVersion not in [1, 2]:
            config.error(f"apiVersion {apiVersion} is not supported")

    @defer.inlineCallbacks
    def reconfigService(self, vaultServer=None, vaultToken=None, secretsmount=None,
                        apiVersion=1):
        if secretsmount is None:
            self.secretsmount = "secret"
        else:
            self.secretsmount = secretsmount
        self.vaultServer = vaultServer
        self.vaultToken = vaultToken
        self.apiVersion = apiVersion
        if vaultServer.endswith('/'):
            vaultServer = vaultServer[:-1]
        self._http = yield httpclientservice.HTTPClientService.getService(
            self.master, self.vaultServer, headers={'X-Vault-Token': self.vaultToken})

    @defer.inlineCallbacks
    def get(self, entry):
        """
        get the value from vault secret backend
        """
        parts = entry.rsplit('/', maxsplit=1)
        name = parts[0]
        if len(parts) > 1:
            key = parts[1]
        else:
            key = 'value'

        if self.apiVersion == 1:
            path = self.secretsmount + '/' + name
        else:
            path = self.secretsmount + '/data/' + name

        # note that the HTTP path contains v1 for both versions of the key-value
        # secret engine. Different versions of the key-value engine are
        # effectively separate secret engines in vault, with the same base HTTP
        # API, but with different paths within it.
        proj = yield self._http.get(f"/v1/{path}")
        code = yield proj.code
        if code != 200:
            raise KeyError(f"The secret {entry} does not exist in Vault provider: request"
                           f" return code: {code}.")
        json = yield proj.json()
        if self.apiVersion == 1:
            secrets = json.get('data', {})
        else:
            secrets = json.get('data', {}).get('data', {})
        try:
            return secrets[key]
        except KeyError as e:
            raise KeyError(
                f"The secret {entry} does not exist in Vault provider: {e}") from e



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import mock

from buildbot.test.steps import TestBuildStepMixin


class SourceStepMixin(TestBuildStepMixin):

    """
    Support for testing source steps.  Aside from the capabilities of
    L{TestBuildStepMixin}, this adds:

     - fake sourcestamps

    The following instance variables are available after C{setupSourceStep}, in
    addition to those made available by L{TestBuildStepMixin}:

    @ivar sourcestamp: fake SourceStamp for the build
    """

    def setUpSourceStep(self):
        return super().setup_test_build_step()

    def tearDownSourceStep(self):
        return super().tear_down_test_build_step()

    # utilities

    def setup_step(self, step, args=None, patch=None, **kwargs):
        """
        Set up C{step} for testing.  This calls L{TestBuildStepMixin}'s C{setup_step}
        and then does setup specific to a Source step.
        """
        step = super().setup_step(step, **kwargs)

        if args is None:
            args = {}

        ss = self.sourcestamp = mock.Mock(name="sourcestamp")
        ss.ssid = 9123
        ss.branch = args.get('branch', None)
        ss.revision = args.get('revision', None)
        ss.project = ''
        ss.repository = ''
        ss.patch = patch
        ss.patch_info = None
        ss.changes = []
        self.build.getSourceStamp = lambda x=None: ss
        return step



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from buildbot.util import ComparableMixin
from buildbot.util import NotABranch
from buildbot.util.ssfilter import _create_branch_filters
from buildbot.util.ssfilter import _create_filters
from buildbot.util.ssfilter import _create_property_filters


class ChangeFilter(ComparableMixin):

    # NOTE: If users use a filter_fn, we have no way to determine whether it has
    # changed at reconfig, so the scheduler will always be restarted.  That's as
    # good as Python can do.
    compare_attrs = (
        'filter_fn',
        'filters',
        'property_filters',
        'project_fn',
        'repository_fn',
        'branch_fn',
        'category_fn',
        'codebase_fn',
    )

    def __init__(
        self,
        filter_fn=None,  # gets a Change object, returns boolean
        project=None,
        project_not_eq=None,
        project_re=None,
        project_not_re=None,
        project_fn=None,
        repository=None,
        repository_not_eq=None,
        repository_re=None,
        repository_not_re=None,
        repository_fn=None,
        branch=NotABranch,
        branch_not_eq=NotABranch,
        branch_re=None,
        branch_not_re=None,
        branch_fn=None,
        category=None,
        category_not_eq=None,
        category_re=None,
        category_not_re=None,
        category_fn=None,
        codebase=None,
        codebase_not_eq=None,
        codebase_re=None,
        codebase_not_re=None,
        codebase_fn=None,
        property_eq=None,
        property_not_eq=None,
        property_re=None,
        property_not_re=None
    ):
        self.filter_fn = filter_fn
        self.project_fn = project_fn
        self.repository_fn = repository_fn
        self.branch_fn = branch_fn
        self.category_fn = category_fn
        self.codebase_fn = codebase_fn

        self.filters = _create_filters(
            project,
            project_not_eq,
            project_re,
            project_not_re,
            'project'
        )
        self.filters += _create_filters(
            repository,
            repository_not_eq,
            repository_re,
            repository_not_re,
            'repository',
        )
        self.filters += _create_branch_filters(
            branch,
            branch_not_eq,
            branch_re,
            branch_not_re,
            'branch',
        )
        self.filters += _create_filters(
            category,
            category_not_eq,
            category_re,
            category_not_re,
            'category',
        )
        self.filters += _create_filters(
            codebase,
            codebase_not_eq,
            codebase_re,
            codebase_not_re,
            'codebase',
        )
        self.property_filters = _create_property_filters(
            property_eq,
            property_not_eq,
            property_re,
            property_not_re,
            'property'
        )

    def filter_change(self, change):
        if self.filter_fn is not None and not self.filter_fn(change):
            return False
        if self.project_fn is not None and not self.project_fn(change.project):
            return False
        if self.codebase_fn is not None and not self.codebase_fn(change.codebase):
            return False
        if self.repository_fn is not None and not self.repository_fn(change.repository):
            return False
        if self.category_fn is not None and not self.category_fn(change.category):
            return False
        if self.branch_fn is not None and not self.branch_fn(change.branch):
            return False

        for filter in self.filters:
            value = getattr(change, filter.prop, '')
            if not filter.is_matched(value):
                return False
        for filter in self.property_filters:
            value = change.properties.getProperty(filter.prop, '')
            if not filter.is_matched(value):
                return False
        return True

    def _get_repr_filters(self):
        filters = []
        if self.filter_fn is not None:
            filters.append(f'{self.filter_fn.__name__}()')
        if self.project_fn is not None:
            filters.append(f'{self.project_fn.__name__}(project)')
        if self.codebase_fn is not None:
            filters.append(f'{self.codebase_fn.__name__}(codebase)')
        if self.repository_fn is not None:
            filters.append(f'{self.repository_fn.__name__}(repository)')
        if self.category_fn is not None:
            filters.append(f'{self.category_fn.__name__}(category)')
        if self.branch_fn is not None:
            filters.append(f'{self.branch_fn.__name__}(branch)')
        filters += [filter.describe() for filter in self.filters]
        filters += [filter.describe() for filter in self.property_filters]
        return filters

    def __repr__(self):
        return f"<{self.__class__.__name__} on {' and '.join(self._get_repr_filters())}>"

    @staticmethod
    def fromSchedulerConstructorArgs(change_filter=None,
                                     branch=NotABranch, categories=None):
        """
        Static method to create a filter based on constructor args
        change_filter, branch, and categories; use default values @code{None},
        @code{NotABranch}, and @code{None}, respectively.  These arguments are
        interpreted as documented for the
        L{buildbot.schedulers.basic.Scheduler} class.

        @returns: L{ChangeFilter} instance or None for not filtering
        """

        # use a change_filter, if given one
        if change_filter:
            if (branch is not NotABranch or categories is not None):
                raise RuntimeError("cannot specify both change_filter and "
                                   "branch or categories")
            return change_filter
        elif branch is not NotABranch or categories:
            # build a change filter from the deprecated category and branch
            # args
            cfargs = {}
            if branch is not NotABranch:
                cfargs['branch'] = branch
            if categories:
                cfargs['category'] = categories
            return ChangeFilter(**cfargs)
        else:
            return None



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


# copy some exceptions from the DB layer
from buildbot.db.schedulers import SchedulerAlreadyClaimedError

__all__ = [
    'SchedulerAlreadyClaimedError',
    'InvalidPathError',
    'InvalidControlException',
]


class DataException(Exception):
    pass


class InvalidPathError(DataException):
    "A path argument was invalid or unknown"


class InvalidControlException(DataException):
    "Action is not supported"


class InvalidQueryParameter(DataException):
    "Query Parameter was invalid"



#!/usr/bin/env python
#
# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

try:
    from buildbot_pkg import setup_www_plugin
except ImportError:
    import sys
    print('Please install buildbot_pkg module in order to install that '
          'package, or use the pre-build .whl modules available on pypi',
          file=sys.stderr)
    sys.exit(1)

setup_www_plugin(
    name='buildbot-codeparameter',
    description='Buildbot Forcescheduler Parameter that use ace.js to display code',
    author=u'Pierre Tardy',
    author_email=u'tardyp@gmail.com',
    url='http://buildbot.net/',
    packages=['buildbot_codeparameter'],
    package_data={
        '': [
            'VERSION',
            'static/*'
        ]
    },
    entry_points="""
        [buildbot.www]
        codeparameter = buildbot_codeparameter:ep
    """,
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)'
    ],
)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest.mock import patch

from parameterized import parameterized

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.secrets.providers.vault_hvac import HashiCorpVaultKvSecretProvider
from buildbot.secrets.providers.vault_hvac import VaultAuthenticatorApprole
from buildbot.secrets.providers.vault_hvac import VaultAuthenticatorToken
from buildbot.test.util import interfaces
from buildbot.test.util.config import ConfigErrorsMixin

try:
    import hvac
    assert hvac
except ImportError:
    hvac = None


class FakeHvacApprole:
    def login(self, role_id, secret_id):
        self.role_id = role_id
        self.secret_id = secret_id


class FakeHvacAuth:
    approle = FakeHvacApprole()


class FakeHvacKvV1:
    token = None

    def read_secret(self, path, mount_point):
        if self.token is None:
            raise hvac.exceptions.Unauthorized
        if path == "wrong/path":
            raise hvac.exceptions.InvalidPath(message="Fake InvalidPath exception")
        return {'data': {'key': "value"}}


class FakeHvacKvV2:
    token = None

    def read_secret_version(self, path, mount_point):
        if self.token is None:
            raise hvac.exceptions.Unauthorized(message="Fake Unauthorized exception")
        if path == "wrong/path":
            raise hvac.exceptions.InvalidPath(message="Fake InvalidPath exception")
        return {'data': {'data': {'key': "value"}}}


class FakeHvacKv:
    default_kv_version = 2
    v1 = FakeHvacKvV1()
    v2 = FakeHvacKvV2()


class FakeHvacSecrets:
    kv = FakeHvacKv()


class FakeHvacClient:
    auth = FakeHvacAuth()
    secrets = FakeHvacSecrets()

    _token = None

    @property
    def token(self):
        return self._token

    @token.setter
    def token(self, new_token):
        self._token = new_token
        self.secrets.kv.v1.token = new_token
        self.secrets.kv.v2.token = new_token

    def is_authenticated(self):
        return self._token


def mock_vault(*args, **kwargs):
    client = FakeHvacClient()
    client.token = "mockToken"
    return client


class TestSecretInVaultAuthenticator(interfaces.InterfaceTests):

    def test_authenticate(self):
        raise NotImplementedError


class TestSecretInVaultAuthenticatorToken(unittest.TestCase, TestSecretInVaultAuthenticator):

    def setUp(self):
        if hvac is None:
            raise unittest.SkipTest(
                "Need to install hvac to test VaultAuthenticatorToken")

    def test_authenticate(self):
        token = "mockToken"
        authenticator = VaultAuthenticatorToken(token)
        client = hvac.Client()
        authenticator.authenticate(client)
        self.assertEqual(client.token, token)


class TestSecretInVaultAuthenticatorApprole(unittest.TestCase, TestSecretInVaultAuthenticator):

    def test_authenticate(self):
        authenticator = VaultAuthenticatorApprole("testRole", "testSecret")
        client = FakeHvacClient()
        authenticator.authenticate(client)
        self.assertEqual(client.auth.approle.secret_id, "testSecret")


class TestSecretInHashiCorpVaultKvSecretProvider(ConfigErrorsMixin, unittest.TestCase):

    def setUp(self):
        if hvac is None:
            raise unittest.SkipTest(
                "Need to install hvac to test HashiCorpVaultKvSecretProvider")
        param = dict(vault_server="", authenticator=VaultAuthenticatorToken("mockToken"),
                     path_delimiter='|', path_escape='\\', api_version=2)
        self.provider = HashiCorpVaultKvSecretProvider(**param)
        self.provider.reconfigService(**param)
        self.provider.client = FakeHvacClient()
        self.provider.client.secrets.kv.default_kv_version = param['api_version']
        self.provider.client.token = "mockToken"

    @parameterized.expand([
        ('vault_server_not_string', {'vault_server': {}}, 'vault_server must be a string'),
        ('path_delimiter_not_char', {'vault_server': 'abc', 'path_delimiter': {}},
         'path_delimiter must be a single character'),
        ('path_delimiter_too_long', {'vault_server': 'abc', 'path_delimiter': 'ab'},
         'path_delimiter must be a single character'),
        ('path_escape_not_char', {'vault_server': 'abc', 'path_escape': {}},
         'path_escape must be a single character'),
        ('path_escape_too_long', {'vault_server': 'abc', 'path_escape': 'ab'},
         'path_escape must be a single character'),
        ('api_version_unsupported', {'vault_server': 'abc', 'api_version': 3},
         'api_version 3 is not supported'),
    ])
    def test_check_config(self, name, params, error):
        with self.assertRaisesConfigError(error):
            HashiCorpVaultKvSecretProvider(authenticator=VaultAuthenticatorToken("mockToken"),
                                           **params)

    def test_check_config_authenticator(self):
        with self.assertRaisesConfigError('authenticator must be instance of VaultAuthenticator'):
            HashiCorpVaultKvSecretProvider(vault_server='abc')

    def test_escaped_split(self):
        parts = self.provider.escaped_split("a/b\\|c/d|e/f\\|g/h")
        self.assertEqual(parts, ["a/b|c/d", "e/f|g/h"])

    def test_escaped_split_ends_with_escape(self):
        parts = self.provider.escaped_split("a|b\\")
        self.assertEqual(parts, ["a", "b"])

    def test_thd_hvac_wrap_read_v1(self):
        self.provider.api_version = 1
        self.provider.client.token = "mockToken"
        value = self.provider.thd_hvac_wrap_read("some/path")
        self.assertEqual(value['data']['key'], "value")

    def test_thd_hvac_wrap_read_v2(self):
        self.provider.client.token = "mockToken"
        value = self.provider.thd_hvac_wrap_read("some/path")
        self.assertEqual(value['data']['data']['key'], "value")

    # for some reason, errors regarding generator function were thrown
    @patch("hvac.Client", side_effect=mock_vault)
    def test_thd_hvac_wrap_read_unauthorized(self, mock_vault):
        self.provider.client.token = None
        yield self.assertFailure(self.provider.thd_hvac_wrap_read("some/path"),
                                 hvac.exceptions.Unauthorized)

    def test_thd_hvac_get_reauthorize(self):
        """
        When token is None, provider gets unauthorized exception and is forced to re-authenticate
        """
        self.provider.client.token = None
        value = self.provider.thd_hvac_get("some/path")
        self.assertEqual(value['data']['data']['key'], "value")

    @defer.inlineCallbacks
    def test_get_v1(self):
        self.provider.api_version = 1
        self.provider.client.token = "mockToken"
        value = yield self.provider.get("some/path|key")
        self.assertEqual(value, "value")

    @defer.inlineCallbacks
    def test_get_v2(self):
        self.provider.client.token = "mockToken"
        value = yield self.provider.get("some/path|key")
        self.assertEqual(value, "value")

    @defer.inlineCallbacks
    def test_get_fail_no_key(self):
        self.provider.client.token = "mockToken"
        with self.assertRaises(KeyError):
            yield self.provider.get("some/path")

    @defer.inlineCallbacks
    def test_get_fail_wrong_key(self):
        self.provider.client.token = "mockToken"
        with self.assertRaises(KeyError):
            yield self.provider.get("some/path|wrong_key")

    @defer.inlineCallbacks
    def test_get_fail_multiple_separators(self):
        self.provider.client.token = "mockToken"
        with self.assertRaises(KeyError):
            yield self.provider.get("some/path|unescaped|key")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
password store based provider
"""

import os
from pathlib import Path

from twisted.internet import defer

from buildbot import config
from buildbot.secrets.providers.base import SecretProviderBase
from buildbot.util import runprocess


class SecretInPass(SecretProviderBase):
    """
    secret is stored in a password store
    """
    name = "SecretInPass"

    def checkPassIsInPath(self):
        if not any((Path(p) / "pass").is_file() for p in os.environ["PATH"].split(":")):
            config.error("pass does not exist in PATH")

    def checkPassDirectoryIsAvailableAndReadable(self, dirname):
        if not os.access(dirname, os.F_OK):
            config.error(f"directory {dirname} does not exist")

    def checkConfig(self, gpgPassphrase=None, dirname=None):
        self.checkPassIsInPath()
        if dirname:
            self.checkPassDirectoryIsAvailableAndReadable(dirname)

    def reconfigService(self, gpgPassphrase=None, dirname=None):
        self._env = {**os.environ}
        if gpgPassphrase:
            self._env["PASSWORD_STORE_GPG_OPTS"] = f"--passphrase {gpgPassphrase}"
        if dirname:
            self._env["PASSWORD_STORE_DIR"] = dirname

    @defer.inlineCallbacks
    def get(self, entry):
        """
        get the value from pass identified by 'entry'
        """
        try:
            rc, output = yield runprocess.run_process(self.master.reactor,
                                                      ['pass', entry], env=self._env,
                                                      collect_stderr=False, stderr_is_error=True)
            if rc != 0:
                return None
            return output.decode("utf-8", "ignore").splitlines()[0]
        except IOError:
            return None



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.process import results
from buildbot.steps import gitdiffinfo
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import TestBuildStepMixin

try:
    import unidiff
except ImportError:
    unidiff = None


class TestDiffInfo(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    if not unidiff:
        skip = 'unidiff is required for GitDiffInfo tests'

    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def tearDown(self):
        return self.tear_down_test_build_step()

    def test_merge_base_failure(self):
        self.setup_step(gitdiffinfo.GitDiffInfo())
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['git', 'merge-base', 'HEAD', 'master'])
            .log('stdio-merge-base', stderr='fatal: Not a valid object name')
            .exit(128))
        self.expect_log_file_stderr('stdio-merge-base', 'fatal: Not a valid object name\n')
        self.expect_outcome(result=results.FAILURE, state_string="GitDiffInfo (failure)")
        return self.run_step()

    def test_diff_failure(self):
        self.setup_step(gitdiffinfo.GitDiffInfo())
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['git', 'merge-base', 'HEAD', 'master'])
            .log('stdio-merge-base', stdout='1234123412341234')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['git', 'diff', '--no-prefix', '-U0', '1234123412341234', 'HEAD'])
            .log('stdio-diff', stderr='fatal: ambiguous argument')
            .exit(1),
            )
        self.expect_log_file('stdio-merge-base', '1234123412341234')
        self.expect_log_file_stderr('stdio-diff', 'fatal: ambiguous argument\n')
        self.expect_outcome(result=results.FAILURE, state_string="GitDiffInfo (failure)")
        return self.run_step()

    def test_empty_diff(self):
        self.setup_step(gitdiffinfo.GitDiffInfo())
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['git', 'merge-base', 'HEAD', 'master'])
            .log('stdio-merge-base', stdout='1234123412341234')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['git', 'diff', '--no-prefix', '-U0', '1234123412341234', 'HEAD'])
            .log('stdio-diff', stdout='')
            .exit(0),
            )
        self.expect_log_file('stdio-merge-base', '1234123412341234')
        self.expect_log_file_stderr('stdio-diff', '')
        self.expect_outcome(result=results.SUCCESS, state_string="GitDiffInfo")
        self.expect_build_data('diffinfo-master', b'[]', 'GitDiffInfo')
        return self.run_step()

    def test_complex_diff(self):
        self.setup_step(gitdiffinfo.GitDiffInfo())
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['git', 'merge-base', 'HEAD', 'master'])
            .log('stdio-merge-base', stdout='1234123412341234')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['git', 'diff', '--no-prefix', '-U0', '1234123412341234', 'HEAD'])
            .log('stdio-diff', stdout='''\
diff --git file1 file1
deleted file mode 100644
index 42f90fd..0000000
--- file1
+++ /dev/null
@@ -1,3 +0,0 @@
-line11
-line12
-line13
diff --git file2 file2
index c337bf1..1cb02b9 100644
--- file2
+++ file2
@@ -4,0 +5,3 @@ line24
+line24n
+line24n2
+line24n3
@@ -15,0 +19,3 @@ line215
+line215n
+line215n2
+line215n3
diff --git file3 file3
new file mode 100644
index 0000000..632e269
--- /dev/null
+++ file3
@@ -0,0 +1,3 @@
+line31
+line32
+line33
''')
            .exit(0)
            )
        self.expect_log_file('stdio-merge-base', '1234123412341234')
        self.expect_outcome(result=results.SUCCESS, state_string="GitDiffInfo")

        diff_info = (
            b'[{"source_file": "file1", "target_file": "/dev/null", ' +
            b'"is_binary": false, "is_rename": false, ' +
            b'"hunks": [{"ss": 1, "sl": 3, "ts": 0, "tl": 0}]}, ' +
            b'{"source_file": "file2", "target_file": "file2", ' +
            b'"is_binary": false, "is_rename": false, ' +
            b'"hunks": [{"ss": 4, "sl": 0, "ts": 5, "tl": 3}, ' +
            b'{"ss": 15, "sl": 0, "ts": 19, "tl": 3}]}, ' +
            b'{"source_file": "/dev/null", "target_file": "file3", ' +
            b'"is_binary": false, "is_rename": false, ' +
            b'"hunks": [{"ss": 0, "sl": 0, "ts": 1, "tl": 3}]}]')
        self.expect_build_data('diffinfo-master', diff_info, 'GitDiffInfo')
        return self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import time

from buildbot_worker.scripts import base
from buildbot_worker.util import rewrap


class Follower(object):

    def follow(self):
        from twisted.internet import reactor
        from buildbot_worker.scripts.logwatcher import LogWatcher
        self.rc = 0
        print("Following twistd.log until startup finished..")
        lw = LogWatcher("twistd.log")
        d = lw.start()
        d.addCallbacks(self._success, self._failure)
        reactor.run()
        return self.rc

    def _success(self, processtype):
        from twisted.internet import reactor
        print("The {0} appears to have (re)started correctly.".format(processtype))
        self.rc = 0
        reactor.stop()

    def _failure(self, why):
        from twisted.internet import reactor
        from buildbot_worker.scripts.logwatcher import WorkerTimeoutError
        if why.check(WorkerTimeoutError):
            print(rewrap("""\
                The worker took more than 10 seconds to start and/or connect
                to the buildmaster, so we were unable to confirm that it
                started and connected correctly.
                Please 'tail twistd.log' and look for a line that says
                'message from master: attached' to verify correct startup.
                If you see a bunch of messages like 'will retry in 6 seconds',
                your worker might not have the correct hostname or portnumber
                for the buildmaster, or the buildmaster might not be running.
                If you see messages like
                   'Failure: twisted.cred.error.UnauthorizedLogin'
                then your worker might be using the wrong botname or password.
                Please correct these problems and then restart the worker.
                """))
        else:
            print(rewrap("""\
                Unable to confirm that the worker started correctly.
                You may need to stop it, fix the config file, and restart.
                """))
            print(why)
        self.rc = 1
        reactor.stop()


def startCommand(config):
    basedir = config['basedir']
    if not base.isWorkerDir(basedir):
        return 1

    return startWorker(basedir, config['quiet'], config['nodaemon'])


def startWorker(basedir, quiet, nodaemon):
    """
    Start worker process.

    Fork and start twisted application described in basedir buildbot.tac file.
    Print it's log messages to stdout for a while and try to figure out if
    start was successful.

    If quiet or nodaemon parameters are True, or we are running on a win32
    system, will not fork and log will not be printed to stdout.

    @param  basedir: worker's basedir path
    @param    quiet: don't display startup log messages
    @param nodaemon: don't daemonize (stay in foreground)
    @return: 0 if worker was successfully started,
             1 if we are not sure that worker started successfully
    """

    os.chdir(basedir)
    if quiet or nodaemon:
        return launch(nodaemon)

    # we probably can't do this os.fork under windows
    from twisted.python.runtime import platformType
    if platformType == "win32":
        return launch(nodaemon)

    # fork a child to launch the daemon, while the parent process tails the
    # logfile
    if os.fork():
        # this is the parent
        rc = Follower().follow()
        return rc
    # this is the child: give the logfile-watching parent a chance to start
    # watching it before we start the daemon
    time.sleep(0.2)
    launch(nodaemon)
    return None


def launch(nodaemon):
    sys.path.insert(0, os.path.abspath(os.getcwd()))

    # see if we can launch the application without actually having to
    # spawn twistd, since spawning processes correctly is a real hassle
    # on windows.
    from twisted.python.runtime import platformType
    from twisted.scripts.twistd import run
    argv = ["twistd",
            "--no_save",
            "--logfile=twistd.log",  # windows doesn't use the same default
            "--python=buildbot.tac"]
    if nodaemon:
        argv.extend(["--nodaemon"])
        if platformType != 'win32':
            # windows doesn't use pidfile option.
            argv.extend(["--pidfile="])

    sys.argv = argv
    run()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer

from buildbot.test.fakedb.base import FakeDBComponent
from buildbot.test.fakedb.row import Row


class BuildData(Row):
    table = 'build_data'

    id_column = 'id'
    foreignKeys = ('buildid',)
    required_columns = ('buildid', 'name', 'value', 'length', 'source')
    binary_columns = ('value',)

    def __init__(self, id=None, buildid=None, name=None, value=None, source=None):
        super().__init__(id=id, buildid=buildid, name=name, value=value, source=source,
                         length=len(value))


class FakeBuildDataComponent(FakeDBComponent):

    def setUp(self):
        self.build_data = {}

    def insertTestData(self, rows):
        for row in rows:
            if isinstance(row, BuildData):
                self.build_data[row.id] = row.values.copy()

    def _get_build_data_row(self, buildid, name):
        for row in self.build_data.values():
            if row['buildid'] == buildid and row['name'] == name:
                return row
        return None

    def setBuildData(self, buildid, name, value, source):
        assert isinstance(value, bytes)
        row = self._get_build_data_row(buildid, name)
        if row is not None:
            row['value'] = value
            row['length'] = len(value)
            row['source'] = source
            return

        id = Row.nextId()
        self.build_data[id] = {
            'id': id,
            'buildid': buildid,
            'name': name,
            'value': value,
            'length': len(value),
            'source': source
        }

    # returns a Deferred
    def getBuildData(self, buildid, name):
        row = self._get_build_data_row(buildid, name)
        if row is not None:
            return defer.succeed(self._row2dict(row))
        return defer.succeed(None)

    # returns a Deferred
    def getBuildDataNoValue(self, buildid, name):
        row = self._get_build_data_row(buildid, name)
        if row is not None:
            return defer.succeed(self._row2dict_novalue(row))
        return defer.succeed(None)

    # returns a Deferred
    def getAllBuildDataNoValues(self, buildid):
        ret = []
        for row in self.build_data.values():
            if row['buildid'] != buildid:
                continue
            ret.append(self._row2dict_novalue(row))

        return defer.succeed(ret)

    # returns a Deferred
    def deleteOldBuildData(self, older_than_timestamp):
        buildids_to_keep = []
        for build_dict in self.db.builds.builds.values():
            if build_dict['complete_at'] is None or \
                    build_dict['complete_at'] >= older_than_timestamp:
                buildids_to_keep.append(build_dict['id'])

        count_before = len(self.build_data)

        build_dataids_to_remove = []
        for build_datadict in self.build_data.values():
            if build_datadict['buildid'] not in buildids_to_keep:
                build_dataids_to_remove.append(build_datadict['id'])

        for id in build_dataids_to_remove:
            self.build_data.pop(id)

        count_after = len(self.build_data)

        return defer.succeed(count_before - count_after)

    def _row2dict(self, row):
        ret = row.copy()
        del ret['id']
        return ret

    def _row2dict_novalue(self, row):
        ret = row.copy()
        del ret['id']
        ret['value'] = None
        return ret



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import absolute_import
from __future__ import print_function

import mock

from twisted.trial import unittest

from buildbot_worker.scripts import restart
from buildbot_worker.scripts import start
from buildbot_worker.scripts import stop
from buildbot_worker.test.util import misc


class TestRestart(misc.IsWorkerDirMixin,
                  misc.StdoutAssertionsMixin,
                  unittest.TestCase):

    """
    Test buildbot_worker.scripts.restart.restart()
    """
    config = {"basedir": "dummy", "nodaemon": False, "quiet": False}

    def setUp(self):
        self.setUpStdoutAssertions()

        # patch start.startWorker() to do nothing
        self.startWorker = mock.Mock()
        self.patch(start, "startWorker", self.startWorker)

    def test_bad_basedir(self):
        """
        test calling restart() with invalid basedir path
        """

        # patch isWorkerDir() to fail
        self.setupUpIsWorkerDir(False)

        # call startCommand() and check that correct exit code is returned
        self.assertEqual(restart.restart(self.config), 1,
                         "unexpected exit code")

        # check that isWorkerDir was called with correct argument
        self.isWorkerDir.assert_called_once_with(self.config["basedir"])

    def test_no_worker_running(self):
        """
        test calling restart() when no worker is running
        """
        # patch basedir check to always succeed
        self.setupUpIsWorkerDir(True)

        # patch stopWorker() to raise an exception
        mock_stopWorker = mock.Mock(side_effect=stop.WorkerNotRunning())
        self.patch(stop, "stopWorker", mock_stopWorker)

        # check that restart() calls startWorker() and outputs correct messages
        restart.restart(self.config)
        self.startWorker.assert_called_once_with(self.config["basedir"],
                                                 self.config["quiet"],
                                                 self.config["nodaemon"])

        self.assertStdoutEqual("no old worker process found to stop\n"
                               "now restarting worker process..\n")

    def test_restart(self):
        """
        test calling restart() when worker is running
        """
        # patch basedir check to always succeed
        self.setupUpIsWorkerDir(True)

        # patch stopWorker() to do nothing
        mock_stopWorker = mock.Mock()
        self.patch(stop, "stopWorker", mock_stopWorker)

        # check that restart() calls startWorker() and outputs correct messages
        restart.restart(self.config)
        self.startWorker.assert_called_once_with(self.config["basedir"],
                                                 self.config["quiet"],
                                                 self.config["nodaemon"])
        self.assertStdoutEqual("now restarting worker process..\n")
